from uuid import UUID

import click

from montecarlodata.collector.validation import CollectorValidationService
from montecarlodata.common.echo_utils import echo_success_message
from montecarlodata.config import Config
from montecarlodata.errors import manage_errors, prompt_connection
from montecarlodata.integrations.onboarding.base import BaseOnboardingService
from montecarlodata.integrations.onboarding.fields import (
    DREMIO_DATABASE_TYPE,
    MYSQL_DB_TYPE,
    ORACLE_DB_TYPE,
    TERADATA_DB_TYPE,
    TRANSACTIONAL_CONNECTION_TYPE,
    TRANSACTIONAL_WAREHOUSE_TYPE,
)


class TransactionalOnboardingService(CollectorValidationService, BaseOnboardingService):
    def __init__(self, config: Config, command_name: str, **kwargs):
        super().__init__(config, command_name=command_name, **kwargs)

    @manage_errors
    def onboard_transactional_db(self, **kwargs) -> None:
        """
        Onboard a Transactional DB connection by validating and adding a connection.
        """
        ssl_options = self.load_ssl_options(options=kwargs)
        if ssl_options:
            kwargs.setdefault("connection_settings", {})["ssl_options"] = ssl_options

        kwargs["connection_type"] = TRANSACTIONAL_CONNECTION_TYPE
        kwargs["warehouse_type"] = TRANSACTIONAL_WAREHOUSE_TYPE

        # These dbTypes have been promoted to their own connection type,
        # instead of being a 'transactional-db' connection type
        promoted_subtypes = [TERADATA_DB_TYPE, DREMIO_DATABASE_TYPE, ORACLE_DB_TYPE, MYSQL_DB_TYPE]
        if kwargs.get("dbType") in promoted_subtypes:
            kwargs["connection_type"] = kwargs.get("dbType")
            kwargs["warehouse_type"] = kwargs.get("dbType")

        # Finds all v2 validations for this connection and runs each one.
        # If they all pass, return a temp credentials key
        # If the --skip-validations flag has been used, no validations are run and just a
        # temp key is returned
        key = self.test_new_credentials(**kwargs)

        # Add the connection
        if key:
            self.add_connection(key, **kwargs)
        else:
            click.Abort()

    @manage_errors
    def update_transactional_db(
        self,
        connection_id: UUID,
        skip_validation: bool,
        validate_only: bool,
        auto_yes: bool,
        connection_type: str,
        **kwargs,
    ) -> None:
        """
        Update a Transactional DB connection by validating and updating a connection.
        """
        ssl_options = self.load_ssl_options(options=kwargs)

        # Remove empty or null inputs. Also remove all ssl_* keys as these
        # will be passed under the 'ssl_options' dict.
        changes = {
            key: value
            for key, value in kwargs.items()
            if value is not None and value != {} and not key.startswith("ssl_")
        }

        if ssl_options:
            changes["ssl_options"] = ssl_options

        # Create a new temp credentials key for this connection with the changes
        key = self.create_update_credentials(
            changes=changes, connection_id=connection_id, connection_type=connection_type
        )

        # Run all validations for this connection using the new temp credentials key
        if not skip_validation:
            self.test_new_credentials_on_existing_connection(
                temp_key=key, connection_id=connection_id, connection_type=connection_type
            )

        # Update the connection with the new credentials key
        if not validate_only:
            prompt_connection(
                message="Validations passed! Would you like to continue?", skip_prompt=auto_yes
            )
            self.update_existing_connection(connection_id=connection_id, temp_key=key)
        else:
            echo_success_message(message="Validations passed!")



import textwrap
from typing import Callable, Dict, List, Optional

import click
from pycarlo.core import Client, Query
from tabulate import tabulate

from montecarlodata.errors import manage_errors


class MonitorService:
    MONITORS_HEADERS = (
        "Monitor UUID",
        "Type",
        "Namespace",
        "Description",
        "Last Update Time",
    )
    MORE_MONITOR_MESSAGE = "There are more monitors available. Increase the limit to view them."
    PSEUDO_MONITOR_TYPE_CB_COMPATIBLE = "CIRCUIT_BREAKER_COMPATIBLE"
    MONITOR_TYPE_STATS = "STATS"
    MONITOR_TYPE_CATEGORIES = "CATEGORIES"
    MONITOR_TYPE_JSON_SCHEMA = "JSON_SCHEMA"
    MONITOR_TYPE_CUSTOM_SQL = "CUSTOM_SQL"
    MONITOR_TYPE_TABLE_METRIC = "TABLE_METRIC"  # Legacy Volume SLIs
    MONITOR_TYPE_FRESHNESS = "FRESHNESS"
    MONITOR_TYPE_VOLUME = "VOLUME"
    MONITOR_TYPES = [
        PSEUDO_MONITOR_TYPE_CB_COMPATIBLE,
        MONITOR_TYPE_CUSTOM_SQL,
        MONITOR_TYPE_TABLE_METRIC,
        MONITOR_TYPE_FRESHNESS,
        MONITOR_TYPE_VOLUME,
        MONITOR_TYPE_STATS,
        MONITOR_TYPE_CATEGORIES,
        MONITOR_TYPE_JSON_SCHEMA,
    ]

    def __init__(
        self,
        client: Client,
        command_name: str,
        print_func: Callable = click.echo,
    ):
        self._pycarlo_client = client
        self._command_name = command_name
        self._print_func = print_func

    @manage_errors
    def list_monitors(
        self,
        limit=100,
        namespaces: Optional[List[str]] = None,
        monitor_types: Optional[List[str]] = None,
    ):
        """
        Get all monitors filter by namespaces and monitor_types
        """
        kwargs: Dict = {"limit": limit + 1}
        if namespaces:
            kwargs["namespaces"] = namespaces
        if monitor_types:
            kwargs["monitor_types"] = monitor_types
        query = Query()
        query.get_monitors(**kwargs).__fields__(
            "uuid",
            "monitor_type",
            "namespace",
            "description",
            "last_update_time",
        )
        response = self._pycarlo_client(
            query,
            additional_headers={
                "x-mcd-telemetry-reason": "cli",
                "x-mcd-telemetry-service": "monitor_service",
                "x-mcd-telemetry-command": self._command_name,
            },
        )
        monitors = [
            (
                item.uuid,
                item.monitor_type,
                item.namespace,
                textwrap.fill(
                    item.description,  # type: ignore
                    width=70,
                )
                if item.description
                else "",
                item.last_update_time,
            )
            for item in response.get_monitors
        ]
        more_ns_available = False
        if len(monitors) > limit:
            monitors = monitors[:-1]
            more_ns_available = True
        self._print_func(tabulate(monitors, headers=self.MONITORS_HEADERS, tablefmt="fancy_grid"))

        if more_ns_available:
            self._print_func(self.MORE_MONITOR_MESSAGE)



import copy
import csv
import dataclasses
import json
import secrets
import time
from collections import OrderedDict
from typing import Dict, Iterable, List, Optional, Union, cast

import click
from box import Box, BoxList
from tabulate import tabulate

from montecarlodata.agents.agent import (
    LAMBDA_ARN_REGEX,
    LAMBDA_ARN_REGEX_GROUP_ACCOUNT_ID,
)
from montecarlodata.agents.fields import AWS
from montecarlodata.common.common import struct_match
from montecarlodata.common.data import (
    AWSArn,
    BucketRow,
    DcResourceProperties,
    EventProperties,
    ResourceProperties,
)
from montecarlodata.common.user import UserService
from montecarlodata.config import Config
from montecarlodata.errors import complain_and_abort, manage_errors, prompt_connection
from montecarlodata.queries.onboarding import EventOnboardingDataQueries
from montecarlodata.utils import AwsClientWrapper, GqlWrapper


class CloudResourceService:
    MCD_EVENT_TYPE_MD = "metadata"
    MCD_EVENT_TYPE_QL = "query-logs"
    MCD_EVENT_TYPE_FRIENDLY_TO_COLLECTOR_OUTPUT_MAP = {
        MCD_EVENT_TYPE_MD: "MetadataEventQueue",
        MCD_EVENT_TYPE_QL: "QueryLogEventQueue",
    }
    MCD_EVENT_Q_SENDER_ID = "__sender"

    _MCD_TOPIC_NAME_PATTERN = "monte-carlo-data-{}-{}-events-topic"
    _MCD_TOPIC_POLICY_SID_PATTERN = "monte-carlo-data-{}-{}-ID"
    _MCD_EVENT_ID_PATTERN = "{}{}{}-notification"

    _MCD_ROLE_PREFIX = "monte-carlo-integration-role"
    _MCD_POLICY_PREFIX = "monte-carlo-integration-cli-policy"
    _MCD_TAGS = [{"Key": "MonteCarloData", "Value": ""}]
    _LIST_EMR_FRIENDLY_HEADERS = ["Name", "Id", "State", "LogUri"]

    def __init__(
        self,
        config: Config,
        command_name: str,
        user_service: Optional[UserService] = None,
        aws_wrapper: Optional[AwsClientWrapper] = None,
        aws_profile_override: Optional[str] = None,
        aws_region_override: Optional[str] = None,
        request_wrapper: Optional[GqlWrapper] = None,
    ):
        self._abort_on_error = True
        self._profile = aws_profile_override
        self._region = aws_region_override
        self._command_name = command_name
        self._request_wrapper = request_wrapper or GqlWrapper(
            config,
            command_name=self._command_name,
        )

        self._user_service = user_service or UserService(
            config=config,
            request_wrapper=self._request_wrapper,
            command_name=self._command_name,
        )
        self._aws_wrapper = aws_wrapper or AwsClientWrapper(
            profile_name=self._profile, region_name=self._region
        )

    @manage_errors
    def create_role(
        self,
        path_to_policy_doc: str,
        dc_id: Optional[str] = None,
        agent_id: Optional[str] = None,
    ) -> None:
        """
        Creates a DC compatible role from the provided policy doc
        """
        current_time = str(time.time())
        external_id = self._generate_random_token()
        role_name = f"{self._MCD_ROLE_PREFIX}-{current_time}"
        policy_name = f"{self._MCD_POLICY_PREFIX}-{current_time}"

        try:
            policy = json.dumps(self._read_json(path_to_policy_doc))
        except json.decoder.JSONDecodeError as err:
            complain_and_abort(f"The provided policy is not valid JSON - {err}")
        else:
            account_id = self._get_aws_account_id(dc_id=dc_id, agent_id=agent_id)
            trust_policy = self._generate_trust_policy(
                account_id=account_id, external_id=external_id
            )
            role_arn = self._aws_wrapper.create_role(
                role_name=role_name, trust_policy=trust_policy, tags=self._MCD_TAGS
            )
            click.echo(f"Created role with ARN - '{role_arn}' and external id - '{external_id}'.")

            self._aws_wrapper.attach_inline_policy(
                role_name=role_name, policy_name=policy_name, policy_doc=policy
            )
            click.echo("Success! Attached provided policy.")

    def _get_aws_account_id(
        self, dc_id: Optional[str] = None, agent_id: Optional[str] = None
    ) -> str:
        """
        Returns the AWS account id of the collector's agent if collector has a remote AWS agent.
        Otherwise, returns the AWS account id of the collector itself.
        """
        if agent_id and not dc_id:
            agent_dict = self._user_service.get_agent(agent_id=agent_id)
            dc_id = agent_dict["dc_id"]  # type: ignore

        collector = self._user_service.get_collector(dc_id=dc_id)
        if not collector.active:
            complain_and_abort("The collector is not active")
        agent = self._user_service.get_collector_agent(dc_id=dc_id)

        if agent:
            matches = LAMBDA_ARN_REGEX.match(agent.endpoint)
            if agent.platform != AWS:
                complain_and_abort("The collector is configured with a non-AWS agent")
            elif matches:
                try:
                    return matches.group(LAMBDA_ARN_REGEX_GROUP_ACCOUNT_ID)
                except IndexError:
                    pass

        if not collector.customerAwsAccountId:
            complain_and_abort(
                "Missing expected property (customerAwsAccountId). "
                "The collector may not have been deployed before"
            )
        return collector.customerAwsAccountId

    @staticmethod
    def _generate_trust_policy(account_id: str, external_id: str) -> str:
        """
        Generates a DC compatible trust policy
        """
        return json.dumps(
            {
                "Version": "2012-10-17",
                "Statement": [
                    {
                        "Effect": "Allow",
                        "Principal": {"AWS": f"arn:aws:iam::{account_id}:root"},
                        "Action": "sts:AssumeRole",
                        "Condition": {"StringEquals": {"sts:ExternalId": external_id}},
                    }
                ],
            }
        )

    @staticmethod
    def _generate_random_token(length: Optional[int] = 16) -> str:
        """
        Generates a random token (e.g. for the external id)
        """
        return secrets.token_urlsafe(length)

    @staticmethod
    def _read_json(path: str) -> Dict:
        """
        Reads a JSON file from the path.
        """
        with open(path) as file:
            return json.load(file)  # loads for the purpose of validating

    @manage_errors
    def list_emr_clusters(
        self,
        only_log_locations: bool = False,
        created_after: Optional[str] = None,
        states: Optional[List] = None,
        no_grid: bool = False,
        headers: str = "firstrow",
        table_format: str = "fancy_grid",
    ):
        """
        Displays information about EMR cluster (name, id, state and log location).
        If only_log_locations is True,
        displays a deduplicated list of EMR log locations.
        """
        clusters = self._aws_wrapper.get_emr_cluster_details(
            created_after=created_after, states=states
        )
        if only_log_locations:
            self._list_emr_clusters_only_log_locations(clusters)
        elif no_grid:
            self._list_emr_clusters_details_no_wait(clusters)
        else:
            self._list_emr_cluster_details(clusters, headers=headers, table_format=table_format)

    def _list_emr_cluster_details(
        self,
        clusters: Iterable[Dict],
        headers: str = "firstrow",
        table_format: str = "fancy_grid",
    ):
        rows = [self._emr_details_row(cluster) for cluster in clusters if cluster]
        click.echo(
            tabulate(
                [self._LIST_EMR_FRIENDLY_HEADERS] + rows,
                headers=headers,
                tablefmt=table_format,
            )
        )

    def _list_emr_clusters_details_no_wait(self, clusters: Iterable[Dict]):
        """
        Display information row by row, used when it can take long to wait for all data
        to format the table.
        """
        found = False
        for cluster in clusters:
            if not cluster:  # Should never happen.
                continue
            found = True
            # Prints line by line to show progress as the list may be long.
            click.echo(
                tabulate([self._emr_details_row(cluster)], tablefmt="plain", maxcolwidths=100)
            )
        if not found:
            click.echo("No clusters found")

    @staticmethod
    def _list_emr_clusters_only_log_locations(clusters: Iterable[Dict]):
        log_locations = set()
        for cluster in clusters:
            if not cluster:  # Should never happen.
                continue
            log_locations.add(cluster.get("LogUri"))
        if not log_locations:
            click.echo("No clusters found")
        else:
            for location in log_locations:
                click.echo(location)

    @staticmethod
    def _emr_details_row(cluster: Dict) -> List[Optional[str]]:
        return [
            cluster["Name"],
            cluster["Id"],
            cluster["Status"]["State"],
            cluster.get("LogUri"),
        ]

    @manage_errors
    def add_events(self, dc_id: Optional[str] = None, auto_yes: bool = False, **kwargs) -> None:
        """
        Adds events for bucket in a repeatable manner (e.g. can be re-run on error in
        between steps) -
            1. Derives collection and resource related properties
            2. Creates topic or updates topic policy if necessary
            3. Updates (or sets) SQS policy with topic if not already set
            4. Subscribes queue to topic if not already done
            5. Creates notification if not already set

        All these steps are idempotent, so if if a permission or resource already exists it will not
        be duplicated.

        This is done via SDK rather than CloudFormation (CF) as CF does not support updating
        notification settings, SQS policies or topic policies for existing resources without using a
        custom resources (i.e. lambda functions), which is a less flexible variant of the actions
        performed here as it is bound by CF inputs and any steps and would still use an AWS SDK,
        without providing any of the common benefits of IaC. Additionally, cross-region and
        cross-account operations are non-trivial in CF, especially in a generic way without prior
        knowledge of the environment.

        Note - the only newly created resource (i.e. SNS topic) can be created externally and
        passed as an option.
        """
        click.echo("[Gathering collector and resource details]")
        collector = self.get_and_validate_active_collector(dc_id=dc_id)
        event_details = self._initialize_events(collector_props=collector, **kwargs)
        self._create_or_update_event_topic(event_details=event_details, skip_prompts=auto_yes)
        self._update_event_q(event_details=event_details, skip_prompts=auto_yes)
        self._create_topic_subscription(event_details=event_details, skip_prompts=auto_yes)
        self._create_event_notification(event_details=event_details, skip_prompts=auto_yes)
        click.echo("\nProcess complete. Have a nice day!")

    def _initialize_events(
        self,
        event_type: str,
        collector_props: Box,
        collector_aws_profile: Optional[str] = None,
        bucket_aws_profile: Optional[str] = None,
        buckets_filename: Optional[str] = None,
        bucket_name: Optional[str] = None,
        prefix: Optional[str] = None,
        suffix: Optional[str] = None,
        **kwargs,
    ) -> EventProperties:
        """
        Collect (or lookup) any relevant event related properties and initialize any clients.

        Note - modifies collector_props by adding `outputs`.
        """
        collector_arn = AWSArn(id=collector_props.stackArn)
        collector_profile = collector_aws_profile or self._profile
        cust_resources_profile = bucket_aws_profile or self._profile

        # Create a client for collector resources and any customer resources.
        # This is necessary as the collector resources might be in a different
        # account/region than customer resources.
        collection_client = AwsClientWrapper(
            profile_name=collector_profile, region_name=collector_arn.region
        )

        collector_props.outputs = collection_client.get_stack_outputs(
            stack_id=collector_props.stackArn
        )
        event_queue_arn = self._get_relevant_q(event_type, collector_props.outputs)

        resource_properties = self._get_resource_properties(
            cust_resources_profile=cust_resources_profile,
            buckets_filename=buckets_filename,
            bucket_name=bucket_name,
            prefix=prefix,
            suffix=suffix,
            **kwargs,
        )

        return EventProperties(
            event_type=event_type,
            event_queue_arn=event_queue_arn,  # type: ignore (assume this won't be None here)
            collection_region=collector_arn.region,
            collection_account_id=collector_arn.account,
            collection_client=collection_client,
            resource_properties=resource_properties,
        )

    def create_event_topic(self, dc_id: Optional[str] = None, auto_yes: bool = False, **kwargs):
        collector = self.get_and_validate_active_collector(dc_id=dc_id)
        event_details = self._populate_event_details_from_saved_config(collector, **kwargs)
        self._create_or_update_event_topic(event_details=event_details, skip_prompts=auto_yes)

    def create_bucket_side_event_infrastructure(
        self, dc_id: Optional[str] = None, auto_yes: bool = False, **kwargs
    ):
        collector = self.get_and_validate_active_collector(dc_id=dc_id)
        event_details = self._populate_event_details_from_saved_config(collector, **kwargs)
        self._create_topic_subscription(event_details=event_details, skip_prompts=auto_yes)
        self._create_event_notification(event_details=event_details, skip_prompts=auto_yes)

    def update_event_queue(self, dc_id: Optional[str] = None, auto_yes: bool = False, **kwargs):
        collector = self.get_and_validate_active_collector(dc_id=dc_id)
        event_details = self._populate_event_details_from_saved_config(
            collector, collection_client=True, **kwargs
        )
        self._update_event_q(event_details=event_details, skip_prompts=auto_yes)

    def save_initial_event_properties(self, dc_id: Optional[str] = None, **kwargs):
        collector = self.get_and_validate_active_collector(dc_id=dc_id)
        event_details = self._populate_event_details_from_saved_config(
            collector, collection_client=True, **kwargs
        )
        self._save_event_properties(event_details)

    def delete_event_properties(self):
        self._delete_event_properties()

    def _get_clients_for_buckets(self, buckets, cust_resources_profile):
        region_to_client = {}
        bucket_to_client = {}
        temp_client = AwsClientWrapper(profile_name=cust_resources_profile)
        for bucket in buckets:
            bucket_name = bucket["bucket_name"]
            region = temp_client.get_bucket_location(bucket_name=bucket_name)
            if region not in region_to_client:
                region_to_client[region] = AwsClientWrapper(
                    profile_name=cust_resources_profile, region_name=region
                )

            bucket_to_client[bucket_name] = region_to_client[region]

        return bucket_to_client

    def parse_buckets_file(self, buckets_filename: str) -> List[BucketRow]:
        buckets_file = csv.DictReader(open(buckets_filename))
        return [
            BucketRow(
                bucket_name=row["bucket_name"],
                prefix=row.get("prefix"),
                suffix=row.get("suffix"),
            )
            for row in buckets_file
        ]

    def _get_resource_properties(
        self,
        cust_resources_profile: Optional[str] = None,
        buckets_filename: Optional[str] = None,
        bucket_name: Optional[str] = None,
        prefix: Optional[str] = None,
        suffix: Optional[str] = None,
        **kwargs,
    ) -> List[ResourceProperties]:
        if not buckets_filename and not bucket_name:
            return []

        if bucket_name:
            args = {}
            if prefix:
                args["prefix"] = prefix
            if suffix:
                args["suffix"] = suffix
            buckets = [{"bucket_name": bucket_name, **args}]

        if buckets_filename:
            bucket_rows = self.parse_buckets_file(buckets_filename)
            buckets = [dataclasses.asdict(bucket_row) for bucket_row in bucket_rows]

        bucket_clients = self._get_clients_for_buckets(buckets, cust_resources_profile)

        return [
            ResourceProperties(
                bucket_name=bucket["bucket_name"],
                client=bucket_clients[bucket["bucket_name"]],
                region=bucket_clients[bucket["bucket_name"]].region,
                account_id=bucket_clients[bucket["bucket_name"]].get_caller_identity(),
                prefix=bucket.get("prefix"),
                suffix=bucket.get("suffix"),
                **kwargs,
            )
            for bucket in buckets
        ]

    def _populate_event_details_from_saved_config(
        self,
        collector_props: Box,
        event_type: Optional[str] = None,
        collector_aws_profile: Optional[str] = None,
        bucket_aws_profile: Optional[str] = None,
        collection_client: bool = False,
        buckets_filename: Optional[str] = None,
        bucket_name: Optional[str] = None,
        prefix: Optional[str] = None,
        suffix: Optional[str] = None,
        **kwargs,
    ):
        collector_arn = AWSArn(id=collector_props.stackArn)
        collector_profile = collector_aws_profile or self._profile
        cust_resources_profile = bucket_aws_profile or self._profile

        response = self._request_wrapper.make_request_v2(
            query=EventOnboardingDataQueries.get.query,
            operation=EventOnboardingDataQueries.get.operation,
            service="event_management",
        )

        config = {"resource_properties": []}
        if response.data:
            config = json.loads(response.data.config)  # type: ignore
            if not collection_client:
                bucket_clients = self._get_clients_for_buckets(
                    config["resource_properties"], cust_resources_profile
                )
            config["resource_properties"] = [
                ResourceProperties(
                    bucket_name=bucket["bucket_name"],
                    client=(
                        bucket_clients[bucket["bucket_name"]] if not collection_client else None
                    ),
                    region=bucket.get("region"),
                    account_id=bucket.get("account_id"),
                    prefix=bucket.get("prefix"),
                    suffix=bucket.get("suffix"),
                    topic_arn=bucket.get("topic_arn"),
                    topic_name=bucket.get("topic_name"),
                )
                for bucket in config["resource_properties"]
            ]

        if not event_type and (config_event_type := config.get("event_type")):
            event_type = cast(str, config_event_type)

        resource_properties = self._get_resource_properties(
            cust_resources_profile=cust_resources_profile,
            buckets_filename=buckets_filename,
            bucket_name=bucket_name,
            prefix=prefix,
            suffix=suffix,
            **kwargs,
        )

        if collection_client:
            aws_client = AwsClientWrapper(
                profile_name=collector_profile, region_name=collector_arn.region
            )
            collector_props.outputs = aws_client.get_stack_outputs(
                stack_id=collector_props.stackArn
            )
            aws_options = dict(
                collection_account_id=aws_client.get_caller_identity(),
                collection_client=aws_client,
                event_queue_arn=config.get("event_queue_arn")
                or self._get_relevant_q(
                    event_type,  # type: ignore
                    collector_props.outputs,
                ),
            )

        else:
            aws_options = dict(
                collection_client=None,
                collection_account_id=config.get("collection_account_id"),
            )

        overrides = {
            "event_type": event_type,
            "resource_properties": resource_properties,
            "collection_region": collector_arn.region,
        }

        event_args = {
            **config,
            **aws_options,
            **{key: value for key, value in overrides.items() if value},
        }

        return EventProperties(
            **event_args,
        )

    def _save_event_properties(self, event_properties: EventProperties):
        config = event_properties.serializable()
        self._request_wrapper.make_request_v2(
            query=EventOnboardingDataQueries.save.query,
            operation=EventOnboardingDataQueries.save.operation,
            service="event_management",
            variables={"config": json.dumps(config)},
        )

    def _delete_event_properties(self):
        self._request_wrapper.make_request_v2(
            query=EventOnboardingDataQueries.delete.query,
            service="event_management",
            operation=EventOnboardingDataQueries.delete.operation,
        )

    def _create_or_update_event_topic(
        self, event_details: EventProperties, skip_prompts: bool = False
    ) -> None:
        """
        Creates topic (if necessary) and/or updates policy to allow access from bucket
        (put) & event queue (subscribe)
        """
        click.echo("[Creating or updating topic properties]")
        q_topic_policy_id = self._MCD_TOPIC_POLICY_SID_PATTERN.format(
            event_details.event_type, event_details.event_queue_name
        )

        region_to_resources = {}
        for resource in event_details.resource_properties:
            region = resource.region
            if region not in region_to_resources:
                region_to_resources[region] = []

            region_to_resources[region].append(resource)

        for region, resources in region_to_resources.items():
            new_statements = []
            update_statements = {}
            region_bucket_arns = [resource.bucket_arn for resource in resources]

            # Create an event topic if an MC one does not already exist in the region
            # for the queue type
            topic_arn = resources[0].topic_arn
            if topic_arn is None:
                name = self._MCD_TOPIC_NAME_PATTERN.format(region, event_details.event_type)
                prompt = (
                    f"Create topic '{name}' in '{region}' "
                    f"({resources[0].account_id})? This is an idempotent operation."
                )
                prompt_connection(message=prompt, skip_prompt=skip_prompts)
                topic_arn = resources[0].client.create_sns_topic(name=name)

            for resource in resources:
                resource.topic_arn = topic_arn
                resource.topic_name = self._validate_topic_arn(resource_details=resource).resource

            topic_attributes = resources[0].client.get_topic_attributes(arn=resource.topic_arn)

            region_bucket_topic_policy_id = self._MCD_TOPIC_POLICY_SID_PATTERN.format(
                event_details.event_type, resources[0].topic_name
            )

            existing_bucket_policy = self._get_preexisting_topic_policy(
                topic_policy_id=region_bucket_topic_policy_id,
                topic_attributes=topic_attributes,
            )

            if not existing_bucket_policy:
                new_statements.append(
                    self._generate_sns_bucket_policy(
                        sid=region_bucket_topic_policy_id,
                        topic_arn=topic_arn,
                        bucket_arns=region_bucket_arns,
                    )
                )
            else:
                update_statements[region_bucket_topic_policy_id] = self._extend_sns_bucket_policy(
                    existing_policy=existing_bucket_policy,
                    bucket_arns=region_bucket_arns,
                )

            if not self._is_preexisting_topic_policy(
                topic_policy_id=q_topic_policy_id, topic_attributes=topic_attributes
            ):
                new_statements.append(
                    self._generate_sns_q_policy(
                        sid=q_topic_policy_id,
                        event_details=event_details,
                        topic_arn=topic_arn,
                    )
                )

            # Update policy if necessary

            new_statement_count = len(new_statements) + len(update_statements)
            if new_statement_count > 0:
                click.echo(
                    f"Adding statements (x{new_statement_count}) to '{resources[0].topic_name}' -"
                )
                click.echo(json.dumps([*new_statements, *update_statements.values()], indent=4))
                prompt_connection(message="Please confirm", skip_prompt=skip_prompts)
                resources[0].client.set_topic_attributes(
                    arn=topic_arn,
                    name="Policy",
                    value=self._extend_topic_policy(
                        topic_attributes=topic_attributes,
                        new_statements=new_statements,
                        update_statements=update_statements,
                    ),
                )
            else:
                click.echo(f"No policy changes required for '{resources[0].topic_name}'")

        self._save_event_properties(event_details)

    def _update_event_q(
        self,
        event_details: EventProperties,
        skip_prompts: bool = False,
    ) -> None:
        """
        Create or update queue policy with topic
        """
        click.echo("[Updating event queue properties]")
        attributes = (
            event_details.collection_client.get_q_attributes(
                name=event_details.event_queue_name, attributes=["Policy"]
            )
            or {}
        )

        topic_arns = list(
            OrderedDict.fromkeys(
                [resource.topic_arn for resource in event_details.resource_properties]
            ).keys()
        )

        if not attributes.get("Attributes", {}).get("Policy"):
            # No policy set (e.g. is a new q). Create
            new_policy = self._generate_q_policy(
                event_details=event_details,
                topic_arns=topic_arns,  # type: ignore (assume all topic arns are valid)
            )
            click.echo(f"No existing policy found. Adding -\n{json.dumps(new_policy, indent=4)}")
        else:
            # Update existing if necessary
            existing_policy = Box(json.loads(attributes.Attributes.Policy))  # type: ignore
            new_policy = self._extend_q_policy(
                existing_policy=existing_policy,
                event_details=event_details,
                topic_arns=topic_arns,  # type: ignore (assume all topic arns are valid)
            )
            if struct_match(
                existing_policy,
                new_policy,  # type: ignore (assume this won't be None here)
            ):
                click.echo(f"Policy for '{event_details.event_queue_name}' requires no changes")
                return
            click.echo(f"Updating existing policy into -\n{json.dumps(new_policy, indent=4)}")

        prompt_connection(message="Please confirm", skip_prompt=skip_prompts)
        event_details.collection_client.set_q_attributes(
            name=event_details.event_queue_name,
            attributes=dict(Policy=json.dumps(new_policy)),
        )

        self._save_event_properties(event_details)

    def _create_event_notification(
        self, event_details: EventProperties, skip_prompts: bool = False
    ) -> None:
        """
        Creates event notification if one does not already exist for the topic
        """
        click.echo("[Creating event notification]")
        for resource in event_details.resource_properties:
            setting_id = self._MCD_EVENT_ID_PATTERN.format(
                AWSArn(id=resource.topic_arn).resource.lower(),  # type: ignore
                f"-{resource.prefix}" if resource.prefix else "",
                f"-{resource.suffix}" if resource.suffix else "",
            )

            # Get existing notification settings for bucket as AWS does not support inserting
            # (only a put operation)
            existing_config = resource.client.get_bucket_event_config(name=resource.bucket_name)
            existing_config.pop("ResponseMetadata", None)
            message = (
                "Updating existing notifications into -"
                if existing_config
                else "Creating notification -"
            )

            new_config = self._merge_event_notifications(
                setting_id=setting_id,
                existing_config=existing_config,
                resource_details=resource,
            )

            if struct_match(existing_config, new_config):
                click.echo(f"Notifications for '{resource.bucket_name}' require no changes")
                continue

            click.echo(f"{message}\n{json.dumps(new_config, indent=4)}")
            prompt_connection(message="Please confirm", skip_prompt=skip_prompts)
            resource.client.set_bucket_event_config(
                name=resource.bucket_name, notification_config=new_config
            )

        self._save_event_properties(event_details)

    def _create_topic_subscription(
        self, event_details: EventProperties, skip_prompts: bool = False
    ) -> None:
        """
        Create topic subscription for endpoint if one for the endpoint already does not exist
        """
        click.echo("[Creating topic subscription]")
        for resource in event_details.resource_properties:
            should_continue = False
            subscriptions = list(
                resource.client.list_topic_subscriptions(arn=resource.topic_arn) or []
            )
            if subscriptions:
                for subscription in [sub for sublist in subscriptions for sub in sublist]:
                    if subscription.get("Endpoint") == event_details.event_queue_arn:
                        click.echo(
                            f"Subscription for '{event_details.event_queue_name}' already exists"
                        )
                        should_continue = True
            if should_continue:
                continue
            click.echo(
                f"Creating subscription for '{event_details.event_queue_name}'"
                f" on '{resource.topic_name}'"
            )
            prompt_connection(message="Please confirm", skip_prompt=skip_prompts)
            resource.client.subscribe_to_topic(
                arn=resource.topic_arn,
                endpoint=event_details.event_queue_arn,
                attributes=dict(RawMessageDelivery="true"),
            )

        self._save_event_properties(event_details)

    def _get_relevant_q(self, event_type: str, outputs: Union[Box, BoxList]) -> None:
        """
        Get queue arn from stack outputs based on the friendly key.

        Raise error if not found (e.g. events are not enabled)
        """
        if (
            not event_type
            or event_type not in self.MCD_EVENT_TYPE_FRIENDLY_TO_COLLECTOR_OUTPUT_MAP.keys()
        ):
            complain_and_abort(f"Event type was not defined or is invalid '{event_type}'.")

        for output in outputs:
            if output.OutputKey == self.MCD_EVENT_TYPE_FRIENDLY_TO_COLLECTOR_OUTPUT_MAP[event_type]:
                return output.OutputValue
        complain_and_abort(
            "Failed to find the relevant event queue. Are events enabled for this collector?"
        )

    @staticmethod
    def _validate_topic_arn(resource_details: ResourceProperties) -> AWSArn:
        """
        Sanity check ARN (e.g. if provided) is valid (format + region)
        """
        try:
            arn = AWSArn(
                id=resource_details.topic_arn,  # type: ignore
            )
        except IndexError:
            complain_and_abort(f"SNS topic ARN '{resource_details.topic_arn}' is not a valid.")
        else:
            if resource_details.region != arn.region:
                complain_and_abort(
                    f"Topic region ({arn.region}) does not match bucket "
                    f"region ({resource_details.region})"
                )
        return arn

    @staticmethod
    def _get_preexisting_topic_policy(topic_policy_id: str, topic_attributes: Box) -> Optional[Box]:
        for statement in Box(json.loads(topic_attributes.Attributes.Policy)).Statement:
            if statement.get("Sid") and statement.Sid == topic_policy_id:
                return statement
        return None

    def _is_preexisting_topic_policy(self, topic_policy_id: str, topic_attributes: Box) -> bool:
        """
        Check if a policy for this bucket already exists. Returns true if it does.
        """
        return bool(self._get_preexisting_topic_policy(topic_policy_id, topic_attributes))

    @staticmethod
    def _extend_topic_policy(
        topic_attributes: Box, new_statements: List[Dict], update_statements: Dict
    ) -> str:
        """
        Merge new statements with existing
        """
        policy = Box(json.loads(topic_attributes.Attributes.Policy))
        new_policy_statement = []
        for statement in policy.Statement:
            if statement.get("Sid") and statement.Sid in update_statements:
                new_policy_statement.append(update_statements[statement.Sid])
            else:
                new_policy_statement.append(statement)
        new_policy_statement.extend(new_statements)
        policy.Statement = new_policy_statement
        return json.dumps(policy)

    @staticmethod
    def _extend_sns_bucket_policy(existing_policy: Box, bucket_arns: List[str]) -> Optional[Dict]:
        """
        Merge new statements with existing
        """
        new_policy = Box(copy.deepcopy(existing_policy))
        source_arn = (
            new_policy.get("Condition", {}).get("StringEquals", {}).get("aws:SourceArn", [])
        )
        if source_arn:
            # Source ARN can be a list or string :/
            if not isinstance(source_arn, List):
                source_arn = [source_arn]

            source_arn = list(OrderedDict.fromkeys([*source_arn, *bucket_arns]).keys())

            new_policy.Condition.StringEquals["aws:SourceArn"] = source_arn
            if len(source_arn) == 1:
                new_policy.Condition.StringEquals["aws:SourceArn"] = source_arn[0]
            return new_policy

        # Existing policy was create outside of tooling / different from steps outlined
        # in documentation
        complain_and_abort("Queue policy exists, but not in a supported format")
        return None

    @staticmethod
    def _extend_q_policy(
        existing_policy: Box,
        event_details: EventProperties,
        topic_arns: Optional[Iterable[str]],
    ) -> Optional[Dict]:
        """
        Merge new statements with existing
        """
        new_policy = Box(copy.deepcopy(existing_policy))
        for statement in new_policy.Statement:
            if statement.get("Sid") and statement.Sid == CloudResourceService.MCD_EVENT_Q_SENDER_ID:
                source_arn = statement.get("Condition", {}).get("ArnLike", {}).get("aws:SourceArn")
                if source_arn:
                    # Source ARN can be a list or string :/
                    list_topic_arns = list(topic_arns)  # type: ignore
                    if isinstance(source_arn, List):
                        source_arn.extend(list_topic_arns)
                    else:
                        source_arn = [source_arn, *list_topic_arns]
                    # Remove duplicates, while still retaining order/type for computing diff.
                    statement.Condition.ArnLike["aws:SourceArn"] = list(
                        OrderedDict.fromkeys(source_arn).keys()
                    )
                    if len(statement.Condition.ArnLike["aws:SourceArn"]) == 1:
                        statement.Condition.ArnLike["aws:SourceArn"] = statement.Condition.ArnLike[
                            "aws:SourceArn"
                        ][0]
                    return new_policy
        # Existing policy was create outside of tooling / different from steps
        # outlined in documentation
        complain_and_abort("Queue policy exists, but not in a supported format")
        return None

    def _merge_event_notifications(
        self,
        setting_id: str,
        existing_config: Dict,
        resource_details: ResourceProperties,
    ) -> Dict:
        """
        Merge new config with existing based on if and how topic config was set
        """
        config = copy.deepcopy(existing_config)
        if config.get("TopicConfigurations"):
            # existing topic notification configurations
            for topic in config["TopicConfigurations"]:
                if topic.get("Id") == setting_id:
                    return config  # No changes required; can short circuit
            config["TopicConfigurations"].append(
                self._generate_bucket_to_topic_config(
                    setting_id=setting_id, resource_details=resource_details
                )
            )
        else:
            # existing queue or lambda notification configurations
            config["TopicConfigurations"] = [
                self._generate_bucket_to_topic_config(
                    setting_id=setting_id, resource_details=resource_details
                )
            ]
        return config

    @staticmethod
    def _generate_bucket_to_topic_config(
        setting_id: str, resource_details: ResourceProperties
    ) -> Dict:
        return dict(
            Id=setting_id,
            TopicArn=resource_details.topic_arn,
            Events=["s3:ObjectCreated:*", "s3:ObjectRemoved:*"],
            Filter=dict(
                Key=dict(
                    FilterRules=[
                        dict(Name="Prefix", Value=resource_details.prefix or ""),
                        dict(Name="Suffix", Value=resource_details.suffix or ""),
                    ]
                )
            ),
        )

    @staticmethod
    def _generate_sns_bucket_policy(sid: str, topic_arn: str, bucket_arns: List[str]) -> Dict:
        return {
            "Sid": sid,
            "Effect": "Allow",
            "Principal": {"AWS": "*"},
            "Action": "SNS:Publish",
            "Resource": topic_arn,
            "Condition": {"StringEquals": {"aws:SourceArn": bucket_arns}},
        }

    @staticmethod
    def _generate_sns_q_policy(sid: str, event_details: EventProperties, topic_arn: str) -> Dict:
        return {
            "Sid": sid,
            "Effect": "Allow",
            "Principal": {"AWS": event_details.collection_account_id},
            "Action": "sns:Subscribe",
            "Resource": topic_arn,
        }

    @staticmethod
    def _generate_q_policy(event_details: EventProperties, topic_arns: Optional[Iterable[str]]):
        return {
            "Version": "2008-10-17",
            "Statement": list(
                filter(
                    None,
                    [
                        {
                            "Sid": "__owner",
                            "Effect": "Allow",
                            "Principal": {
                                "AWS": f"arn:aws:iam::{event_details.collection_account_id}:root"
                            },
                            "Action": "SQS:*",
                            "Resource": event_details.event_queue_arn,
                        },
                        (
                            {
                                "Sid": CloudResourceService.MCD_EVENT_Q_SENDER_ID,
                                "Effect": "Allow",
                                "Principal": {"AWS": "*"},
                                "Action": "SQS:SendMessage",
                                "Resource": event_details.event_queue_arn,
                                "Condition": {"ArnLike": {"aws:SourceArn": list(topic_arns)}},
                            }
                            if topic_arns
                            else None
                        ),
                    ],
                )
            ),
        }

    def get_and_validate_active_collector(
        self, dc_id: Optional[str] = None, agent_id: Optional[str] = None
    ) -> Box:
        """
        Get a specific collector and validate that it is active.
        """
        collector = self._user_service.get_collector(dc_id=dc_id, agent_id=agent_id)
        if not collector.stackArn or not collector.active:
            complain_and_abort("Cannot setup events for an inactive collector.")
        return collector

    def get_dc_resource_props(
        self,
        *,
        collector_props: Box,
        collector_aws_profile: Optional[str] = None,
        resource_aws_profile: Optional[str] = None,
        resource_aws_region: Optional[str] = None,
        get_stack_outputs: bool = True,
        get_stack_params: bool = True,
        **_,
    ) -> DcResourceProperties:
        """
        Collect (or lookup) DC and resource properties. Note - modifies collector_props by
        adding `outputs` and `parameters`.

        Multiple clients are created as collector resources might be in a different
        account/region than customer resources.
        """
        collector_arn = AWSArn(id=collector_props.stackArn)
        collector_profile = collector_aws_profile or self._profile
        cust_resources_profile = resource_aws_profile or self._profile

        collection_region = collector_arn.region
        resources_region = resource_aws_region or collection_region

        collection_client = AwsClientWrapper(
            profile_name=collector_profile, region_name=collection_region
        )
        resources_client = AwsClientWrapper(
            profile_name=cust_resources_profile, region_name=resources_region
        )

        if get_stack_outputs:
            collector_props.outputs = self._parse_stack_prop_list(
                key="OutputKey",
                val="OutputValue",
                struct=collection_client.get_stack_outputs(stack_id=collector_props.stackArn),
            )

        if get_stack_params:
            collector_props.parameters = self._parse_stack_prop_list(
                key="ParameterKey",
                val="ParameterValue",
                struct=collection_client.get_stack_parameters(stack_id=collector_props.stackArn),
            )

        return DcResourceProperties(
            collector_arn=collector_arn,
            collector_props=collector_props,
            collection_region=collection_region,
            resources_region=resources_region,
            collection_client=collection_client,
            resources_client=resources_client,
        )

    @staticmethod
    def _parse_stack_prop_list(key: str, val: str, struct: Union[List[Dict], Box]) -> Box:
        """
        Parse stack props into an easier to use collapsed format
        """
        return Box({row[key]: row[val] for row in struct}, camel_killer_box=True)



import json
from dataclasses import dataclass
from typing import List, Optional

from dataclasses_json import DataClassJsonMixin, LetterCase
from dataclasses_json.cfg import config
from marshmallow import fields

from montecarlodata.iac.utils import field_spec
from montecarlodata.settings import (
    DEFAULT_EXCLUDE_PATTERNS,
    DEFAULT_INCLUDE_PATTERNS,
    DEFAULT_MONTECARLO_MONITOR_CONFIG_VERSION,
)


@dataclass
class ProjectConfig(DataClassJsonMixin):
    version: Optional[int] = field_spec(
        fields.Int(required=False),
        default_factory=lambda: DEFAULT_MONTECARLO_MONITOR_CONFIG_VERSION,
    )
    default_resource: Optional[str] = field_spec(fields.Str(required=False))
    include_file_patterns: Optional[List[str]] = field_spec(
        fields.List(fields.Str(required=True)),
        default_factory=lambda: DEFAULT_INCLUDE_PATTERNS,
    )
    exclude_file_patterns: Optional[List[str]] = field_spec(
        fields.List(fields.Str(required=True)), default_factory=list
    )
    namespace: Optional[str] = None

    def __post_init__(self):
        self.exclude_file_patterns = list(
            set((self.exclude_file_patterns or []) + DEFAULT_EXCLUDE_PATTERNS)
        )


@dataclass
class ResourceModification(DataClassJsonMixin):
    dataclass_json_config = config(letter_case=LetterCase.CAMEL)["dataclasses_json"]

    type: str
    description: str
    is_significant_change: bool = False
    diff_string: Optional[str] = None
    resource_type: Optional[str] = None
    resource_index: Optional[int] = None


@dataclass
class ConfigTemplateUpdateAsyncResponse(DataClassJsonMixin):
    dataclass_json_config = config(letter_case=LetterCase.CAMEL)["dataclasses_json"]

    update_uuid: Optional[str] = None
    errors_as_json: Optional[str] = None
    warnings_as_json: Optional[str] = None

    def __post_init__(self):
        self.errors = json.loads(self.errors_as_json) if self.errors_as_json else {}


@dataclass
class ConfigTemplateUpdateState(DataClassJsonMixin):
    dataclass_json_config = config(letter_case=LetterCase.CAMEL)["dataclasses_json"]

    state: str
    resource_modifications: Optional[List[ResourceModification]] = None
    errors_as_json: Optional[str] = None
    warnings_as_json: Optional[str] = None
    changes_applied: bool = False

    def __post_init__(self):
        self.resource_modifications = self.resource_modifications or []
        self.errors = json.loads(self.errors_as_json) if self.errors_as_json else {}

    @property
    def warnings(self):
        return json.loads(self.warnings_as_json) if self.warnings_as_json else {}


@dataclass
class ConfigTemplateDeleteResponse(DataClassJsonMixin):
    dataclass_json_config = config(letter_case=LetterCase.CAMEL)["dataclasses_json"]

    num_deleted: int
    changes_applied: bool = False



# Query response fields
MATCH_AND_CREATE_BI_WH_SRC_GQL_OPERATION = "matchAndCreateBiWarehouseSources"
MATCH_AND_CREATE_BI_WH_SRC_GQL_RESPONSE_FIELD = "match_and_create_bi_warehouse_sources"
MATCHING_BI_WH_SRC_GQL_RESPONSE_FIELD = "matching_bi_warehouse_sources"
MATCH_SUCCESSFUL_GQL_RESPONSE_FIELD = "match_successful"



from typing import Iterator, List, Optional, Tuple

import click
from tabulate import tabulate

from montecarlodata.common.common import is_overlap
from montecarlodata.common.data import AWSArn, DcResourceProperties
from montecarlodata.common.resources import CloudResourceService
from montecarlodata.errors import complain_and_abort, manage_errors
from montecarlodata.integrations.onboarding.fields import REDSHIFT_CONNECTION_TYPE


class NetworkDiscoveryService:
    MCD_NETWORK_REC_RESOURCE_TYPE_MAP = {REDSHIFT_CONNECTION_TYPE: "rs_network_recommender"}

    def __init__(
        self,
        command_name: str,
        cloud_resource_service: Optional[CloudResourceService] = None,
        *args,
        **kwargs,
    ):
        self._abort_on_error = True  # Used in decorator
        self._cloud_resource_service = cloud_resource_service or CloudResourceService(
            command_name=command_name,
            *args,
            **kwargs,
        )

    @manage_errors
    def recommend_network_dispatcher(
        self, *, resource_type: str, dc_id: Optional[str] = None, **kwargs
    ) -> None:
        """
        Make a network connection recommendation by resource to help simplify steps
        required to onboard.
        """
        try:
            dispatch = self.MCD_NETWORK_REC_RESOURCE_TYPE_MAP[resource_type]
        except KeyError:
            complain_and_abort(f"Unsupported resource type - '{resource_type}'.")
        else:
            collector = self._cloud_resource_service.get_and_validate_active_collector(dc_id=dc_id)
            dc_resource_props = self._cloud_resource_service.get_dc_resource_props(
                collector_props=collector, **kwargs
            )

            getattr(self, dispatch)(dc_resource_props=dc_resource_props, **kwargs)

    def rs_network_recommender(
        self, *, dc_resource_props: DcResourceProperties, resource_identifier: str, **_
    ) -> None:
        """
        Alpha recommender for how to connect the DC with RS.

        Connections can be made with either IP filtering or peering. Peering has the following
        scenarios -
            1. Deployed without a VPC
            2. Overlapping CIDR ranges
            3. Same account peering
            4. Multi account peering

        Multi-region is supported. Transit gateways are supported, but not considered by the
        recommender as MC currently pushes peering.

        https://docs.getmontecarlo.com/docs/aws-networking-troubleshoot-guide#after-data-collector-deployment
        """
        self._check_collector_validity(dc_resource_props)

        # Retrieve properties from the redshift cluster
        dc_resource_props.resource_props = (
            dc_resource_props.resources_client.describe_redshift_cluster(
                cluster_id=resource_identifier
            )
        )
        dc_resource_props.resource_arn = AWSArn(
            dc_resource_props.resource_props.cluster_namespace_arn  # type: ignore
        )

        does_redshift_dc_share_an_account = (
            dc_resource_props.resource_arn.account == dc_resource_props.collector_arn.account
        )
        redshift_security_groups, redshift_in_vpc = self._get_rs_sg(dc_resource_props)

        redshift_vpc_range = dc_resource_props.resources_client.describe_vpc(
            dc_resource_props.resource_props.vpc_id  # type: ignore
        ).cidr_block
        collector_vpc_range = dc_resource_props.collector_props.parameters.vpc_cidr

        base_header = "The redshift cluster '{}' is{}publicly accessible."

        # Build recommendations based on the cluster's availability (public or not).
        if dc_resource_props.resource_props.publicly_accessible:  # type: ignore
            header = base_header.format(resource_identifier, " ")
            recommendations = self._get_ip_filtering_verbiage(
                dc_resource_props, redshift_security_groups
            )
        else:
            if not redshift_in_vpc:
                # Clusters deployed outside of a VPC are essentially a legacy feature and not
                # supported.
                complain_and_abort(
                    "Recommender does not support non-public clusters deployed outside a VPC."
                )

            header = base_header.format(resource_identifier, " not ")
            recommendation_intro = self._get_peering_intro_verbiage(
                does_redshift_dc_share_an_account
            )

            if is_overlap(redshift_vpc_range, collector_vpc_range):
                # This case is considered a recommendation and not an error as is a
                # "supportable path".
                recommendations = [
                    *recommendation_intro,
                    self._get_overlap_verbiage(collector_vpc_range, redshift_vpc_range),
                ]
            else:
                redshift_routes = self._get_rs_routes(dc_resource_props)

                recommendations = [
                    *recommendation_intro,
                    *self._get_rs_cross_account_verbiage(
                        dc_resource_props, does_redshift_dc_share_an_account
                    ),
                    *self._get_rs_requester_verbiage(
                        dc_resource_props,
                        does_redshift_dc_share_an_account,
                        redshift_vpc_range,
                    ),
                    *self._get_rs_accepter_verbiage(
                        dc_resource_props, redshift_security_groups, redshift_routes
                    ),
                    *self._get_extra_verbiage(),
                ]

        self._echo_recommendation([*recommendations], header_col=header)

    @staticmethod
    def _check_collector_validity(
        dc_resource_props: DcResourceProperties,
        default_unset_val: Optional[str] = "N/A",
    ) -> None:
        """
        Error on unsupported collectors (e.g. using an existing VPC).
        """
        if (
            "existing_vpc_id" in dc_resource_props.collector_props.parameters
            and dc_resource_props.collector_props.parameters.existing_vpc_id != default_unset_val
        ):
            complain_and_abort(
                "Recommender does not support collectors deployed in customer managed VPCs."
            )

    @staticmethod
    def _get_rs_sg(
        dc_resource_props: DcResourceProperties,
        expected_status: Optional[str] = "active",
    ) -> Tuple[List[str], bool]:
        """
        Parse the security groups associated with the RS cluster.

        Returns a list of SGs and a flag to indicate the cluster is deployed in a VPC (yes=True).
        """
        if all(
            prop in dc_resource_props.resource_props  # type: ignore
            for prop in ["vpc_security_groups", "cluster_security_groups"]
        ):
            if (
                len(dc_resource_props.resource_props.vpc_security_groups) > 0  # type: ignore
                and len(dc_resource_props.resource_props.cluster_security_groups) > 0  # type: ignore
            ):
                # This should not be possible as AWS does not support this case.
                complain_and_abort(
                    "Recommender does not support clusters with ambiguous hosting. "
                    "Cluster cannot have security groups both inside and outside a VPC."
                )
            elif len(dc_resource_props.resource_props.vpc_security_groups) > 0:  # type: ignore
                return [
                    sg.vpc_security_group_id
                    for sg in dc_resource_props.resource_props.vpc_security_groups  # type: ignore
                    if sg.status.lower() == expected_status
                ], True
            elif len(dc_resource_props.resource_props.cluster_security_groups) > 0:  # type: ignore
                return [
                    sg.cluster_security_group_name
                    for sg in dc_resource_props.resource_props.cluster_security_groups  # type: ignore
                    if sg.status.lower() == expected_status
                ], False
        return [], False

    @staticmethod
    def _get_rs_routes(dc_resource_props: DcResourceProperties) -> Iterator[str]:
        """
        Get routes for the cluster.
        """
        if "cluster_subnet_group_name" not in dc_resource_props.resource_props:  # type: ignore
            complain_and_abort("Recommender does not support clusters deployed outside a VPC.")

        rs_subnets = [
            subnet.subnet_identifier
            for subnet in dc_resource_props.resources_client.describe_cluster_subnet_groups(
                subnet_group=dc_resource_props.resource_props.cluster_subnet_group_name,  # type: ignore
            ).subnets
        ]

        for subnet in rs_subnets or []:
            route = dc_resource_props.resources_client.describe_routes(
                filter_vals=[subnet]
            ).route_tables
            if len(route) == 0:
                route = dc_resource_props.resources_client.describe_routes(
                    filter_vals=["true"], filter_key="association.main"
                ).route_tables
            yield route[0].route_table_id

    @staticmethod
    def _get_ip_filtering_verbiage(
        dc_resource_props: DcResourceProperties, redshift_security_groups: List[str]
    ) -> List[str]:
        """
        Generate verbiage for IP filtering.
        """
        return [
            "IP filtering is recommended. See steps below.",
            "",
            f"Whitelist port '{dc_resource_props.resource_props.endpoint.port}' for "  # type: ignore
            f"'{dc_resource_props.collector_props.outputs.public_ip}/32' in any of the following "
            f"redshift security groups - {redshift_security_groups}.",
            "https://docs.getmontecarlo.com/docs/network-connectivity#ip-filtering",
        ]

    @staticmethod
    def _get_rs_cross_account_verbiage(
        dc_resource_props: DcResourceProperties, does_redshift_dc_share_an_account: bool
    ) -> List[str]:
        """
        Generate verbiage for cross account stack stack. This is an optional step based on
        the accounts used.
        """
        if does_redshift_dc_share_an_account:
            return []
        return [
            "- Deploy the 'Create Peering Cross Account assumable role CloudFormation stack' "
            "with the following values -",
            f"1. Data Collector AWS Account ID: {dc_resource_props.collector_arn.account}",
            "https://docs.getmontecarlo.com/docs/peering-templates#create-peering-cross-account-assumable-role-cloudformation-stack",
            "",
        ]

    @staticmethod
    def _get_rs_requester_verbiage(
        dc_resource_props: DcResourceProperties,
        does_redshift_dc_share_an_account: bool,
        redshift_vpc_range: str,
    ) -> List[str]:
        """
        Generate verbiage for requester stack.
        """
        cross_role_verbiage = []
        if not does_redshift_dc_share_an_account:
            # Only visible if the previous stack was necessary
            cross_role_verbiage = [
                "8. VPC peer role for cross AWS account connections: "
                "<PeeringRole from the output of the previous stack>"
            ]

        return [
            "- Deploy the  'Create Requester CloudFormation stack' with the following values -",
            f"1. Monte Carlo Data Collector VPC ID: '{dc_resource_props.collector_props.outputs.vpc_id}'",  # noqa: E501
            f"2. Monte Carlo Data Collector Security Group ID: '{dc_resource_props.collector_props.outputs.security_group}'",  # noqa: E501
            f"3. Monte Carlo Data Collector Route Table ID: '{dc_resource_props.collector_props.outputs.private_route_table}'",  # noqa: E501
            f"4. Warehouse/resource VPC ID: '{dc_resource_props.resource_props.vpc_id}'",  # type: ignore
            f"5. Warehouse/resource AWS Account ID: '{dc_resource_props.resource_arn.account}'",  # type: ignore
            f"6. Warehouse/resource AWS Region: '{dc_resource_props.resource_arn.region}'",  # type: ignore
            f"7. Warehouse/resource CIDR Block: '{redshift_vpc_range}'",
            *cross_role_verbiage,
            "https://docs.getmontecarlo.com/docs/peering-templates#create-requester-cloudformation-stack",
            "",
        ]

    @staticmethod
    def _get_rs_accepter_verbiage(
        dc_resource_props: DcResourceProperties,
        redshift_security_groups: List[str],
        redshift_routes: Iterator[str],
    ) -> List[str]:
        """
        Generate verbiage for accepter stack.
        """
        return [
            "- Deploy the  'Create Accepter CloudFormation stack' with the following values -",
            f"1. Monte Carlo Data Collector CIDR Block: '{dc_resource_props.collector_props.parameters.vpc_cidr}'",  # noqa: E501
            "2. Monte Carlo Data Collector Peering Connection: <PeeringConnection from the output of the previous stack>",  # noqa: E501
            f"3. Resource / Warehouse Security Group: {NetworkDiscoveryService._get_friendly_sg_verbiage(redshift_security_groups)}",  # noqa: E501
            *[
                f"{count + 3}. Resource / Warehouse Route Table #{count}: '{route}'"
                for count, route in enumerate(set(redshift_routes), start=1)
            ],
            "https://docs.getmontecarlo.com/docs/peering-templates#create-accepter-cloudformation-stack",
            "",
        ]

    @staticmethod
    def _get_overlap_verbiage(collector_vpc_range: str, redshift_vpc_range: str) -> str:
        """
        Generate verbiage for overlaps.
        """
        return (
            f"Collector range ({collector_vpc_range}) overlaps with "
            f"resource ({redshift_vpc_range}). "
            f"Peering is not possible when peered VPCs use overlapping CIDR blocks. "
            f"Please redeploy the collector with a custom CIDR block then rerun recommender."
        )

    @staticmethod
    def _get_peering_intro_verbiage(
        does_redshift_dc_share_an_account: bool,
    ) -> List[str]:
        """
        Generate verbiage for peering header / intro.
        """
        section_count = 2 if does_redshift_dc_share_an_account else 3

        return [
            f"VPC Peering is recommended. See the outlined steps for each "
            f"of the {section_count} sections below.",
            "",
            "The CloudFormation template can be found by following the link at the end of "
            "each section.",
            f"Please complete all {section_count} sections, and for any additional help, "
            "reach out to your Monte Carlo representative.",
            "",
        ]

    @staticmethod
    def _get_extra_verbiage() -> List[str]:
        """
        Generate any extra verbiage (i.e. not path specific).
        """
        return [
            "If the cluster's subnets use a non-default ACL the collector CIDR "
            "Block also likely need to be whitelisted."
        ]

    @staticmethod
    def _get_friendly_sg_verbiage(redshift_security_groups: List[str]) -> str:
        """
        Get Human readable SG groups
        """
        return (
            f"Any of the following - {','.join(redshift_security_groups)}"
            if len(redshift_security_groups) > 1
            else f"'{redshift_security_groups[0]}'"
        )

    @staticmethod
    def _echo_recommendation(
        recs: List[str],
        header_col: str = "Recommendations",
        table_format: str = "pipe",
    ) -> None:
        """
        Display recommendations in a fancy table.
        """
        click.echo(
            tabulate(
                [[header_col]] + [[rec] for rec in recs],
                headers="firstrow",
                tablefmt=table_format,
            )
        )



import base64
from unittest import TestCase
from unittest.mock import Mock, patch

from montecarlodata.common.user import UserService
from montecarlodata.integrations.onboarding.bi.reports import ReportsOnboardingService
from montecarlodata.queries.onboarding import (
    ADD_BI_CONNECTION_MUTATION,
    TEST_LOOKER_GIT_CLONE_CRED_MUTATION,
    TEST_LOOKER_METADATA_CRED_MUTATION,
    TEST_POWER_BI_CRED_MUTATION,
    TEST_TABLEAU_CRED_MUTATION,
)
from montecarlodata.utils import AwsClientWrapper, GqlWrapper
from tests.test_base_onboarding import _SAMPLE_BASE_OPTIONS
from tests.test_common_user import _SAMPLE_CONFIG

_SAMPLE_OPTIONS = {"foo": "bar"}


class ReportOnboardingTest(TestCase):
    def setUp(self) -> None:
        self._user_service_mock = Mock(autospec=UserService)
        self._request_wrapper_mock = Mock(autospec=GqlWrapper)
        self._aws_wrapper_mock = Mock(autospec=AwsClientWrapper)

        self._service = ReportsOnboardingService(
            _SAMPLE_CONFIG,
            command_name="test",
            request_wrapper=self._request_wrapper_mock,
            aws_wrapper=self._aws_wrapper_mock,
            user_service=self._user_service_mock,
        )

    @patch.object(ReportsOnboardingService, "onboard")
    def test_tableau_flow(self, onboard_mock):
        self._service.onboard_tableau(**_SAMPLE_BASE_OPTIONS)
        onboard_mock.assert_called_once_with(
            validation_query=TEST_TABLEAU_CRED_MUTATION,
            connection_query=ADD_BI_CONNECTION_MUTATION,
            validation_response="testTableauCredentials",
            connection_response="addBiConnection",
            connection_type="tableau",
            warehouse_type="tableau",
            **_SAMPLE_BASE_OPTIONS,
        )

    @patch.object(ReportsOnboardingService, "onboard")
    def test_looker_metadata_flow(self, onboard_mock):
        self._service.onboard_looker_metadata(**_SAMPLE_BASE_OPTIONS)
        onboard_mock.assert_called_once_with(
            validation_query=TEST_LOOKER_METADATA_CRED_MUTATION,
            connection_query=ADD_BI_CONNECTION_MUTATION,
            validation_response="testLookerCredentials",
            connection_response="addBiConnection",
            connection_type="looker",
            warehouse_type="looker",
            **_SAMPLE_BASE_OPTIONS,
        )

    @patch.object(ReportsOnboardingService, "onboard")
    @patch("montecarlodata.integrations.onboarding.bi.reports.read_as_base64")
    def test_looker_git_flow(self, read_as_base64_mock, onboard_mock):
        file_path, service_json = "foo", "bar"
        base64_key = base64.b64encode(service_json.encode("utf-8"))

        input_options = {"ssh_key": file_path, **_SAMPLE_BASE_OPTIONS}
        expected_options = {**{"ssh_key": base64_key.decode()}, **_SAMPLE_BASE_OPTIONS}

        read_as_base64_mock.return_value = base64_key

        self._service.onboard_looker_git(**input_options)
        read_as_base64_mock.assert_called_once_with(file_path)
        onboard_mock.assert_called_once_with(
            validation_query=TEST_LOOKER_GIT_CLONE_CRED_MUTATION,
            connection_query=ADD_BI_CONNECTION_MUTATION,
            validation_response="testLookerGitCloneCredentials",
            connection_response="addBiConnection",
            connection_type="looker-git-clone",
            warehouse_type="looker",
            **expected_options,
        )

    @patch.object(ReportsOnboardingService, "onboard")
    def test_power_bi_flow(self, onboard_mock):
        self._service.onboard_power_bi(**_SAMPLE_BASE_OPTIONS)
        onboard_mock.assert_called_once_with(
            validation_query=TEST_POWER_BI_CRED_MUTATION,
            connection_query=ADD_BI_CONNECTION_MUTATION,
            validation_response="testPowerBiCredentials",
            connection_response="addBiConnection",
            connection_type="power-bi",
            warehouse_type="power-bi",
            **_SAMPLE_BASE_OPTIONS,
        )



import os
from pathlib import Path

"""Environmental configuration"""

# Verbose error logging
MCD_VERBOSE_ERRORS = os.getenv("MCD_VERBOSE_ERRORS", False) in (True, "true", "True")

# MCD API endpoint. Overwrites config-file
MCD_API_ENDPOINT = os.getenv("MCD_API_ENDPOINT")

# Default MCD API endpoint when no env or config file setting is available
MCD_DEFAULT_API_ENDPOINT = "https://api.getmontecarlo.com/graphql"

# MCD ID header (for use in local development and testing)
MCD_USER_ID_HEADER = os.getenv("MCD_USER_ID_HEADER")

# MCD API ID. Overwrites config-file
MCD_DEFAULT_API_ID = os.getenv("MCD_DEFAULT_API_ID")

# MCD API Token. Overwrites config-file
MCD_DEFAULT_API_TOKEN = os.getenv("MCD_DEFAULT_API_TOKEN")

# MCD Agent image host. Overwrites config-file
MCD_AGENT_IMAGE_HOST = os.getenv("MCD_AGENT_IMAGE_HOST")

# Default MCD Agent image host when no env or config file setting is available
MCD_DEFAULT_AGENT_IMAGE_HOST = "docker.io"

# MCD Agent image organization. Overwrites config-file
MCD_AGENT_IMAGE_ORG = os.getenv("MCD_AGENT_IMAGE_ORG")

# Default MCD Agent image organization when no env or config file setting is available
MCD_DEFAULT_AGENT_IMAGE_ORG = "montecarlodata"

# MCD Agent image repository. Overwrites config-file
MCD_AGENT_IMAGE_REPO = os.getenv("MCD_AGENT_IMAGE_REPO")

# Default MCD Agent image repository when no env or config file setting is available
MCD_DEFAULT_AGENT_IMAGE_REPO = "agent"

# dbt cloud API token
DBT_CLOUD_API_TOKEN = os.getenv("DBT_CLOUD_API_TOKEN")

# dbt cloud account ID
DBT_CLOUD_ACCOUNT_ID = os.getenv("DBT_CLOUD_ACCOUNT_ID")


"""Tool Defaults"""

# Default profile to be used
DEFAULT_PROFILE_NAME = "default"

# Default path where any configuration files are written
DEFAULT_CONFIG_PATH = os.path.join(str(Path.home()), ".mcd")

# Default region where data collector is deployed
DEFAULT_AWS_REGION = "us-east-1"

"""Internal Use"""

# File name for profile configuration
PROFILE_FILE_NAME = "profiles.ini"

# Configuration sub-command
CONFIG_SUB_COMMAND = "configure"

# Help flag of arguments and options
HELP_FLAG = "--help"

# Option file flag
OPTION_FILE_FLAG = "--option-file"

# Value (if entered) to show prompt with hidden input
SHOW_PROMPT_VALUE = "-1"

"""Monitors as code configs"""

DEFAULT_MONTECARLO_MONITOR_CONFIG_VERSION = 1

PROJECT_CONFIG_FILENAME_YML = "montecarlo.yml"
PROJECT_CONFIG_FILENAME_YAML = "montecarlo.yaml"

DBT_CONFIG_YML = "dbt_project.yml"
DBT_CONFIG_YAML = "dbt_project.yaml"

TARGET_DIRECTORY_NAME = "target"

DEFAULT_INCLUDE_PATTERNS = [
    "**/*.yaml",
    "**/*.yml",
]

DEFAULT_EXCLUDE_PATTERNS = [
    "target/*",
    DBT_CONFIG_YML,
    DBT_CONFIG_YAML,
    PROJECT_CONFIG_FILENAME_YML,
    PROJECT_CONFIG_FILENAME_YAML,
]



import json
import time
from typing import Dict, Optional

from montecarlodata.iac.schemas import (
    ConfigTemplateDeleteResponse,
    ConfigTemplateUpdateAsyncResponse,
    ConfigTemplateUpdateState,
)
from montecarlodata.queries.iac import (
    CREATE_OR_UPDATE_MONTE_CARLO_CONFIG_TEMPLATE_ASYNC,
    DELETE_MONTE_CARLO_CONFIG_TEMPLATE,
    GET_MONTE_CARLO_CONFIG_TEMPLATE_UPDATE_STATE,
)
from montecarlodata.utils import GqlWrapper


class MonteCarloConfigTemplateClient:
    ASYNC_TIMEOUT_SECONDS = 60 * 15 * 2  # 30m (2x the lambda timeout of 15 minutes)

    def __init__(self, gql_wrapper: GqlWrapper):
        self._gql_wrapper = gql_wrapper

    def apply_config_template(
        self,
        namespace: str,
        config_template_as_dict: Dict,
        resource: Optional[str] = None,
        dry_run: bool = False,
        misconfigured_as_warning: bool = False,
        create_non_ingested_tables: bool = False,
    ) -> Optional[ConfigTemplateUpdateState]:
        response = self.apply_config_template_async(
            namespace=namespace,
            config_template_as_dict=config_template_as_dict,
            resource=resource,
            dry_run=dry_run,
            misconfigured_as_warning=misconfigured_as_warning,
            create_non_ingested_tables=create_non_ingested_tables,
        )

        if response.errors:
            return ConfigTemplateUpdateState(
                state="FAILED",
                resource_modifications=[],
                errors_as_json=response.errors_as_json,
                changes_applied=False,
            )
        update_uuid = response.update_uuid
        start = time.time()
        while time.time() - start < self.ASYNC_TIMEOUT_SECONDS:
            state = self.get_config_template_update_state(update_uuid)

            if state.state != "PENDING":
                if response.warnings_as_json:
                    response_warnings = json.loads(response.warnings_as_json)
                    state_warnings = {}
                    if state.warnings_as_json:
                        state_warnings = json.loads(state.warnings_as_json)
                    state_warnings.update(response_warnings)
                    state.warnings_as_json = json.dumps(state_warnings)
                return state

            time.sleep(5)

    def apply_config_template_async(
        self,
        namespace: str,
        config_template_as_dict: Dict,
        resource: Optional[str] = None,
        dry_run: bool = False,
        misconfigured_as_warning: bool = False,
        create_non_ingested_tables: bool = False,
    ) -> ConfigTemplateUpdateAsyncResponse:
        response = self._gql_wrapper.make_request_v2(
            query=CREATE_OR_UPDATE_MONTE_CARLO_CONFIG_TEMPLATE_ASYNC,
            operation="createOrUpdateMonteCarloConfigTemplateAsync",
            service="iac_service",
            variables=dict(
                namespace=namespace,
                configTemplateJson=json.dumps(config_template_as_dict),
                dryRun=dry_run,
                misconfiguredAsWarning=misconfigured_as_warning,
                resource=resource,
                createNonIngestedTables=create_non_ingested_tables,
            ),
        )
        assert response.data is not None
        return ConfigTemplateUpdateAsyncResponse.from_dict(response.data["response"])

    def get_config_template_update_state(
        self,
        update_uuid: Optional[str],
    ) -> ConfigTemplateUpdateState:
        response = self._gql_wrapper.make_request_v2(
            query=GET_MONTE_CARLO_CONFIG_TEMPLATE_UPDATE_STATE,
            operation="getMonteCarloConfigTemplateUpdateState",
            service="iac_service",
            variables=dict(updateUuid=update_uuid),
        )

        return ConfigTemplateUpdateState.from_dict(response.data)

    def delete_config_template(
        self, namespace: str, dry_run: bool = False
    ) -> ConfigTemplateDeleteResponse:
        response = self._gql_wrapper.make_request_v2(
            query=DELETE_MONTE_CARLO_CONFIG_TEMPLATE,
            operation="deleteMonteCarloConfigTemplate",
            service="iac_service",
            variables=dict(
                namespace=namespace,
                dryRun=dry_run,
            ),
        )
        assert response.data is not None
        return ConfigTemplateDeleteResponse.from_dict(response.data["response"])



from montecarlodata.config import Config
from montecarlodata.errors import manage_errors
from montecarlodata.integrations.onboarding.base import BaseOnboardingService
from montecarlodata.integrations.onboarding.fields import (
    EXPECTED_PRESTO_SQL_GQL_RESPONSE_FIELD,
    PRESTO_CERT_PREFIX,
    PRESTO_SQL_CONNECTION_TYPE,
)
from montecarlodata.queries.onboarding import (
    TEST_PRESTO_CRED_MUTATION,
)


class PrestoOnboardingService(BaseOnboardingService):
    def __init__(self, config: Config, command_name: str, **kwargs):
        super().__init__(config, command_name=command_name, **kwargs)

    @manage_errors
    def onboard_presto_sql(self, **kwargs) -> None:
        """
        Onboard a presto-sql connection by validating and adding a connection.
        Also, optionally uploads a certificate to the DC bucket.
        """
        self.handle_cert(cert_prefix=PRESTO_CERT_PREFIX, options=kwargs)
        self.onboard(
            validation_query=TEST_PRESTO_CRED_MUTATION,
            validation_response=EXPECTED_PRESTO_SQL_GQL_RESPONSE_FIELD,
            connection_type=PRESTO_SQL_CONNECTION_TYPE,
            **kwargs,
        )



from dataclasses import dataclass, field
from enum import Enum
from typing import Any, Dict, List, Optional, Set, Union

from box import Box, BoxList
from dataclasses_json import CatchAll, LetterCase, Undefined, dataclass_json


class ConnectionType(Enum):
    """
    Enumeration of connection type values from the GraphQL API
    """

    BigQuery = "BIGQUERY"
    DataLake = "DATA_LAKE"
    Hive = "HIVE"
    Redshift = "REDSHIFT"
    Snowflake = "SNOWFLAKE"


@dataclass
class MonolithResponse:
    data: Optional[Union[Box, Dict, BoxList]] = None
    errors: Optional[List[Dict]] = None


@dataclass_json(letter_case=LetterCase.CAMEL, undefined=Undefined.INCLUDE)  # type: ignore
@dataclass
class OnboardingConfiguration:
    connection_options: CatchAll  # ConnectionOptions

    connection_type: str
    connection_name: Optional[str] = None

    validation_query: Optional[str] = None
    validation_response: Optional[str] = None

    connection_query: Optional[str] = None
    connection_response: Optional[str] = None

    warehouse_type: Optional[str] = None
    warehouse_name: Optional[str] = None

    job_limits: Optional[Dict] = None
    job_types: Optional[List[str]] = None

    connection_id: Optional[str] = None

    create_warehouse: bool = True

    is_active: bool = True
    is_consumer_connection: bool = False


@dataclass_json(letter_case=LetterCase.CAMEL, undefined=Undefined.INCLUDE)  # type: ignore
@dataclass
class ConnectionOptions:
    monolith_base_payload: CatchAll  # Base options passed to the monolith (e.g. host, db, etc.)

    # Client connection options
    dc_id: Optional[str] = None
    validate_only: bool = False
    skip_validation: bool = False
    skip_permission_tests: bool = False
    auto_yes: bool = False

    monolith_connection_payload: dict = field(
        init=False
    )  # Additional options passed to the monolith (from client)

    def __post_init__(self):
        self.monolith_connection_payload = {}

        if self.dc_id:
            self.dc_id = str(self.dc_id)
            self.monolith_connection_payload["dcId"] = self.dc_id
        if self.skip_validation:
            self.monolith_connection_payload["skipValidation"] = self.skip_validation
        if self.skip_permission_tests:
            self.monolith_connection_payload["skipPermissionTests"] = self.skip_permission_tests


@dataclass
class ValidationResult:
    has_warnings: bool
    credentials_key: str


@dataclass
class AWSArn:
    id: str

    # components
    arn: str = field(init=False)
    partition: str = field(init=False)
    service: str = field(init=False)
    region: str = field(init=False)
    account: str = field(init=False)
    resource: str = field(init=False)
    resource_type: Optional[str] = field(init=False)

    def __post_init__(self):
        elements = self.id.split(":", 5)
        self.arn = elements[0]
        self.partition = elements[1]
        self.service = elements[2]
        self.region = elements[3]
        self.account = elements[4]
        self.resource = elements[5]
        self.resource_type = None

        if "/" in self.resource:
            self.resource_type, self.resource = self.resource.split("/", 1)
        elif ":" in self.resource:
            self.resource_type, self.resource = self.resource.split(":", 1)


@dataclass
class BucketRow:
    bucket_name: str
    prefix: Optional[str] = None
    suffix: Optional[str] = None


@dataclass
class ResourceProperties:
    bucket_name: str
    bucket_arn: str = field(init=False)

    region: str
    client: Any
    account_id: str

    prefix: Optional[str] = None
    suffix: Optional[str] = None

    topic_arn: Optional[str] = None
    topic_name: Optional[str] = None

    def __post_init__(self):
        self.bucket_arn = f"arn:aws:s3:::{self.bucket_name}"

    def serializable(self):
        serialize_keys = [
            "bucket_name",
            "region",
            "account_id",
            "prefix",
            "suffix",
            "topic_arn",
            "topic_name",
        ]
        return {key: getattr(self, key) for key in serialize_keys}


@dataclass
class EventProperties:
    event_type: str

    event_queue_arn: str
    event_queue_name: str = field(init=False)

    collection_region: str
    collection_account_id: str
    collection_client: Any

    resource_properties: List[ResourceProperties]

    def __post_init__(self):
        self.event_queue_name = (
            AWSArn(self.event_queue_arn).resource if self.event_queue_arn else None
        )  # type: ignore

    def serializable(self):
        serialize_keys = [
            "event_type",
            "event_queue_arn",
            "collection_region",
            "collection_account_id",
        ]
        serializable_dict = {key: getattr(self, key) for key in serialize_keys}

        serializable_dict["resource_properties"] = [
            resource.serializable() for resource in self.resource_properties
        ]

        return serializable_dict


@dataclass
class DcResourceProperties:
    collector_arn: AWSArn
    collector_props: Box

    collection_region: str
    resources_region: str

    collection_client: Any
    resources_client: Any

    resource_arn: Optional[AWSArn] = None
    resource_props: Optional[Box] = None


@dataclass
class AwsGlueAthenaResourceProperties:
    database_names: Set[str]
    bucket_names: Set[str]

    dc_resource_props: Optional[DcResourceProperties] = None



from unittest import TestCase
from unittest.mock import Mock, patch

from box import Box

from montecarlodata.collector.network_tests import CollectorNetworkTestService
from montecarlodata.common.data import MonolithResponse
from montecarlodata.queries.collector import (
    TEST_TCP_OPEN_CONNECTION,
    TEST_TELNET_CONNECTION,
)
from montecarlodata.utils import GqlWrapper
from tests.test_common_user import _SAMPLE_CONFIG

SAMPLE_OPTIONS = {"loki": "thor"}
SAMPLE_MESSAGE = "Telnet connection for getmontecarlo.com:443 is usable."
SAMPLE_RESPONSE = MonolithResponse(
    data=Box(
        {
            "success": True,
            "validations": [{"type": "network", "message": SAMPLE_MESSAGE}],
        }
    ),
    errors=None,
)


class CollectorNetworkTest(TestCase):
    def setUp(self) -> None:
        self._request_wrapper_mock = Mock(autospec=GqlWrapper)

        self._service = CollectorNetworkTestService(
            _SAMPLE_CONFIG,
            command_name="test",
            request_wrapper=self._request_wrapper_mock,
        )

    @patch("montecarlodata.collector.network_tests.click")
    def test_echo_telnet(self, click_mock):
        self._request_wrapper_mock.make_request_v2.return_value = SAMPLE_RESPONSE
        self._service.echo_telnet_test(**SAMPLE_OPTIONS)
        self._request_wrapper_mock.make_request_v2.assert_called_once_with(
            query=TEST_TELNET_CONNECTION,
            operation="testTelnetConnection",
            service="collector_network_test",
            variables=SAMPLE_OPTIONS,
        )
        click_mock.echo.assert_called_once_with(SAMPLE_MESSAGE)

    @patch("montecarlodata.collector.network_tests.click")
    def test_echo_tcp_open(self, click_mock):
        self._request_wrapper_mock.make_request_v2.return_value = SAMPLE_RESPONSE
        self._service.echo_tcp_open_test()
        self._request_wrapper_mock.make_request_v2.assert_called_once_with(
            query=TEST_TCP_OPEN_CONNECTION,
            operation="testTcpOpenConnection",
            service="collector_network_test",
            variables={},
        )
        click_mock.echo.assert_called_once_with(SAMPLE_MESSAGE)



import click

from montecarlodata.data_exports.data_exports import DataExportService
from montecarlodata.data_exports.fields import (
    ALERTS_DATA_EXPORT_NAME,
    ASSETS_DATA_EXPORT_NAME,
    EVENTS_DATA_EXPORT_NAME,
    FILE_SCHEME,
    MONITORS_DATA_EXPORT_NAME,
    S3_SCHEME,
    SCHEME_DELIM,
)
from montecarlodata.tools import add_common_options

# Shared command verbiage
GET_VERBIAGE = f"""
\b\n
DESTINATION is the path where the data will be written to.

Supported schemes:\n
    '{FILE_SCHEME}{SCHEME_DELIM}' - save Data Export locally.\n
    '{S3_SCHEME}{SCHEME_DELIM}' - save Data Export to S3.

Notice - Will overwrite a file if it exists in the path and create any missing directories or prefixes.
"""  # noqa: E501

# Options shared across commands
MINIMAL_GET_OPTIONS = [
    click.option(
        "--aws-profile",
        required=False,
        help="AWS profile to be used when uploading to S3.",
    ),
    click.option(
        "--dry",
        required=False,
        is_flag=True,
        help="Echo temporary presigned URL for the Data Export and quit.",
    ),
]

GET_OPTIONS = [click.argument("destination", required=True), *MINIMAL_GET_OPTIONS]


@click.group(help="Export your data.")
def export():
    """
    Group for any Data Export related subcommands
    """
    pass


@export.command(help="List Data Export details and availability.", name="list")
@click.pass_obj
def list_data_exports(ctx):
    DataExportService(
        config=ctx["config"],
        command_name="export list_data_exports",
    ).echo_data_exports()


@export.command(help=f"Export monitors data. {GET_VERBIAGE}")
@add_common_options(GET_OPTIONS)
@click.pass_obj
def get_monitors(ctx, **kwargs):
    DataExportService(
        config=ctx["config"],
        command_name="export get_monitors",
    ).get_data_export(data_export=MONITORS_DATA_EXPORT_NAME, **kwargs)


@export.command(help=f"Export assets data. {GET_VERBIAGE}")
@add_common_options(GET_OPTIONS)
@click.pass_obj
def get_assets(ctx, **kwargs):
    DataExportService(
        config=ctx["config"],
        command_name="export get_assets",
    ).get_data_export(data_export=ASSETS_DATA_EXPORT_NAME, **kwargs)


@export.command(help=f"Export alerts data. {GET_VERBIAGE}")
@add_common_options(GET_OPTIONS)
@click.pass_obj
def get_alerts(ctx, **kwargs):
    DataExportService(
        config=ctx["config"],
        command_name="export get_alerts",
    ).get_data_export(data_export=ALERTS_DATA_EXPORT_NAME, **kwargs)


@export.command(help=f"Export events data. {GET_VERBIAGE}")
@add_common_options(GET_OPTIONS)
@click.pass_obj
def get_events(ctx, **kwargs):
    DataExportService(
        config=ctx["config"],
        command_name="export get_events",
    ).get_data_export(data_export=EVENTS_DATA_EXPORT_NAME, **kwargs)


@export.command(help=f"Export Monte Carlo objects. {GET_VERBIAGE}")
@add_common_options(MINIMAL_GET_OPTIONS)
@click.option(
    "--name",
    help="Type of objects to export. Names can be found "
    'via the list command (e.g. `MONITORS` from "Monitors (MONITORS)")',
    required=True,
)
@click.option(
    "--destination",
    help="Destination location to save the data.",
    required=True,
)
@click.pass_obj
def get(ctx, name, **kwargs):
    DataExportService(
        config=ctx["config"],
        command_name="export get",
    ).get_data_export(data_export=name, **kwargs)



from typing import Optional
from unittest import TestCase
from unittest.mock import Mock, call, patch

import click
from box import Box, BoxList

from montecarlodata.common.data import MonolithResponse
from montecarlodata.data_exports.data_exports import DataExportService
from montecarlodata.queries.data_exports import GET_DATA_EXPORT_URL
from montecarlodata.utils import GqlWrapper
from tests.test_common_user import _SAMPLE_CONFIG

SAMPLE_RAW = "loki"
SAMPLE_DATA_EXPORT = "MONITORS"
SAMPLE_DATA_EXPORT_URL = "https://montecarlodata.com"
SAMPLE_S3_DESTINATION = "s3://bucket/prefix/object"
SAMPLE_FS_DESTINATION = "file://folder1/folder2/object"


class MockRequest:
    raw = SAMPLE_RAW


class DataExportServiceTest(TestCase):
    def setUp(self) -> None:
        self._request_wrapper_mock = Mock(autospec=GqlWrapper)
        self._service = DataExportService(
            _SAMPLE_CONFIG,
            command_name="test",
            request_wrapper=self._request_wrapper_mock,
        )

    @patch("montecarlodata.data_exports.data_exports.click")
    def test_echo_data_exports(self, click_mock):
        self._request_wrapper_mock.make_request_v2.return_value = MonolithResponse(
            data=BoxList(
                [
                    {
                        "name": "MONITORS",
                        "title": "Monitors",
                        "description": "All monitors with aggregated properties, excluding deleted monitors.",
                    },
                    {
                        "name": "ASSETS",
                        "title": "Assets",
                        "description": "All assets with aggregated properties, excluding deleted assets.",
                    },
                    {
                        "name": "ALERTS",
                        "title": "Alerts",
                        "description": "All alerts in the last 90 days with aggregated properties.",
                    },
                    {
                        "name": "EVENTS",
                        "title": "Events",
                        "description": "All events in the last 90 days with aggregated properties.",
                    },
                ]
            ),
            errors=None,
        )

        self._service.echo_data_exports(table_format="plain")
        click_mock.echo.assert_called_once_with(
            "Data Export (Name)    Description\n"
            "Monitors (MONITORS)   "
            "All monitors with aggregated properties, excluding deleted monitors.\n"
            "Assets (ASSETS)       "
            "All assets with aggregated properties, excluding deleted assets.\n"
            "Alerts (ALERTS)       "
            "All alerts in the last 90 days with aggregated properties.\n"
            "Events (EVENTS)       "
            "All events in the last 90 days with aggregated properties."
        )

    def test_get_data_export_with_bad_scheme(self):
        with self.assertRaises(click.exceptions.Abort):
            self._service.get_data_export(data_export=SAMPLE_DATA_EXPORT, destination="foo")

        with self.assertRaises(click.exceptions.Abort):
            self._service.get_data_export(data_export=SAMPLE_DATA_EXPORT, destination="http://foo")

    @patch.object(DataExportService, "_save_data_export_to_disk")
    @patch.object(DataExportService, "_save_data_export_to_s3")
    def test_get_data_export_routing_s3(self, save_to_s3_mock, save_to_disk_mock):
        with self.assertRaises(click.exceptions.Abort):
            self._test_get_data_export_routing(destination=SAMPLE_S3_DESTINATION)

        save_to_disk_mock.assert_not_called()
        save_to_s3_mock.assert_not_called()

        self._test_get_data_export_routing(destination=SAMPLE_S3_DESTINATION, aws_profile="foo")

        save_to_disk_mock.assert_not_called()
        save_to_s3_mock.assert_called_once_with(
            data_export_url=SAMPLE_DATA_EXPORT_URL,
            destination="bucket/prefix/object",
            aws_profile="foo",
        )

    @patch.object(DataExportService, "_save_data_export_to_disk")
    @patch.object(DataExportService, "_save_data_export_to_s3")
    def test_get_data_export_routing_fs(self, save_to_s3_mock, save_to_disk_mock):
        self._test_get_data_export_routing(destination=SAMPLE_FS_DESTINATION)

        save_to_disk_mock.assert_called_once_with(
            data_export_url=SAMPLE_DATA_EXPORT_URL,
            destination="folder1/folder2/object",
            aws_profile=None,
        )
        save_to_s3_mock.assert_not_called()

    @patch("montecarlodata.data_exports.data_exports.click")
    @patch.object(DataExportService, "_get_data_export_url")
    @patch.object(DataExportService, "_save_data_export_to_disk")
    @patch.object(DataExportService, "_save_data_export_to_s3")
    def test_get_data_export_dry(self, save_to_s3_mock, save_to_disk_mock, get_mock, mock_click):
        get_mock.return_value = SAMPLE_DATA_EXPORT_URL

        service = DataExportService(
            _SAMPLE_CONFIG,
            command_name="test",
            request_wrapper=self._request_wrapper_mock,
        )
        service.get_data_export(
            data_export=SAMPLE_DATA_EXPORT, destination=SAMPLE_FS_DESTINATION, dry=True
        )
        get_mock.assert_called_once_with(data_export=SAMPLE_DATA_EXPORT)
        mock_click.echo.assert_called_once_with(SAMPLE_DATA_EXPORT_URL)

        save_to_disk_mock.assert_not_called()
        save_to_s3_mock.assert_not_called()

    @patch("montecarlodata.data_exports.data_exports.click")
    @patch.object(DataExportService, "_get_data_export_url")
    def _test_get_data_export_routing(
        self, get_mock, click_mock, destination: str, aws_profile: Optional[str] = None
    ):
        get_mock.return_value = SAMPLE_DATA_EXPORT_URL

        service = DataExportService(
            _SAMPLE_CONFIG,
            command_name="test",
            request_wrapper=self._request_wrapper_mock,
        )
        service.get_data_export(
            data_export=SAMPLE_DATA_EXPORT, destination=destination, aws_profile=aws_profile
        )
        get_mock.assert_called_once_with(data_export=SAMPLE_DATA_EXPORT)
        click_mock.echo.assert_has_calls(
            [
                call(f"Saving data export to '{destination}'."),
                call("Complete. Have a nice day!"),
            ]
        )
        self.assertEqual(click_mock.echo.call_count, 2)

    def test_get_data_export_url(self):
        self._request_wrapper_mock.make_request_v2.return_value = MonolithResponse(
            data=Box({"url": SAMPLE_DATA_EXPORT_URL})
        )
        self.assertEqual(
            self._service._get_data_export_url(data_export=SAMPLE_DATA_EXPORT),
            SAMPLE_DATA_EXPORT_URL,
        )

        self._request_wrapper_mock.make_request_v2.assert_called_once_with(
            query=GET_DATA_EXPORT_URL,
            operation="getDataExportUrl",
            service="data_exports_service",
            variables=dict(data_export_name=SAMPLE_DATA_EXPORT),
        )

    def test_get_data_export_url_not_available(self):
        self._request_wrapper_mock.make_request_v2.return_value = MonolithResponse(
            data=Box({"url": None})
        )

        with self.assertRaises(click.exceptions.Abort):
            self._service._get_data_export_url(data_export=SAMPLE_DATA_EXPORT)

        self._request_wrapper_mock.make_request_v2.assert_called_once_with(
            query=GET_DATA_EXPORT_URL,
            operation="getDataExportUrl",
            service="data_exports_service",
            variables=dict(data_export_name=SAMPLE_DATA_EXPORT),
        )

    @patch("montecarlodata.data_exports.data_exports.urllib")
    @patch("montecarlodata.data_exports.data_exports.mkdirs")
    def test_save_data_export_to_disk(self, mkdir_mock, urllib_mock):
        destination = "folder1/folder2/object"
        self._service._save_data_export_to_disk(
            data_export_url=SAMPLE_DATA_EXPORT_URL, destination=destination
        )
        mkdir_mock.assert_called_once_with("folder1/folder2")
        urllib_mock.request.urlretrieve.assert_called_once_with(SAMPLE_DATA_EXPORT_URL, destination)

    @patch("montecarlodata.data_exports.data_exports.urllib")
    @patch("montecarlodata.data_exports.data_exports.mkdirs")
    def test_save_data_export_to_disk_with_no_destination(self, mkdir_mock, urllib_mock):
        with self.assertRaises(click.exceptions.Abort):
            self._service._save_data_export_to_disk(
                data_export_url=SAMPLE_DATA_EXPORT_URL, destination=None
            )
        mkdir_mock.assert_not_called()
        urllib_mock.request.urlretrieve.assert_not_called()

    @patch("montecarlodata.data_exports.data_exports.AwsClientWrapper")
    @patch("montecarlodata.data_exports.data_exports.requests")
    def test_save_data_export_to_s3(self, request_mock, utils_mock):
        destination = "bucket/prefix/object"
        request_mock.get.return_value = MockRequest

        self._service._save_data_export_to_s3(
            data_export_url=SAMPLE_DATA_EXPORT_URL, destination=destination
        )
        request_mock.get.assert_called_once_with(SAMPLE_DATA_EXPORT_URL, stream=True)
        utils_mock().upload_stream_to_s3.assert_called_once_with(
            data=SAMPLE_RAW, bucket="bucket", key="prefix/object"
        )

    @patch("montecarlodata.data_exports.data_exports.AwsClientWrapper")
    @patch("montecarlodata.data_exports.data_exports.requests")
    def test_save_data_export_to_s3_with_bad_destination(self, request_mock, utils_mock):
        with self.assertRaises(click.exceptions.Abort):
            self._service._save_data_export_to_s3(
                data_export_url=SAMPLE_DATA_EXPORT_URL, destination="bucket"
            )
        request_mock.assert_not_called()
        utils_mock.assert_not_called()

        request_mock.reset_mock()
        utils_mock.reset_mock()

        with self.assertRaises(click.exceptions.Abort):
            self._service._save_data_export_to_s3(
                data_export_url=SAMPLE_DATA_EXPORT_URL, destination="bucket/"
            )
        request_mock.assert_not_called()
        utils_mock.assert_not_called()



from montecarlodata.common.data import AWSArn
from montecarlodata.errors import complain_and_abort, manage_errors
from montecarlodata.integrations.onboarding.base import BaseOnboardingService
from montecarlodata.integrations.onboarding.fields import (
    CONNECTION_TO_WAREHOUSE_TYPE_MAP,
    EXPECTED_ADD_BI_RESPONSE_FIELD,
    EXPECTED_SELF_HOSTED_GQL_RESPONSE_FIELD,
    EXPECTED_SELF_HOSTED_GQL_V2_RESPONSE_FIELD,
    LOOKER_BI_TYPE,
    LOOKER_GIT_CLONE_CONNECTION_TYPE,
    LOOKER_MD_CONNECTION_TYPE,
    POWER_BI_BI_TYPE,
    POWER_BI_CONNECTION_TYPE,
    TABLEAU_BI_TYPE,
    TABLEAU_CONNECTION_TYPE,
)
from montecarlodata.queries.onboarding import (
    ADD_BI_CONNECTION_MUTATION,
    TEST_SELF_HOSTED_CRED_MUTATION,
    TEST_SELF_HOSTED_CRED_V2_MUTATION,
)


class SelfHostedCredentialOnboardingService(BaseOnboardingService):
    BI_CONNECTION_TYPES = {
        LOOKER_GIT_CLONE_CONNECTION_TYPE,
        LOOKER_MD_CONNECTION_TYPE,
        TABLEAU_CONNECTION_TYPE,
        POWER_BI_CONNECTION_TYPE,
    }

    BI_WAREHOUSE_TYPE_MAPPING = {
        TABLEAU_CONNECTION_TYPE: TABLEAU_BI_TYPE,
        POWER_BI_CONNECTION_TYPE: POWER_BI_BI_TYPE,
        LOOKER_MD_CONNECTION_TYPE: LOOKER_BI_TYPE,
        LOOKER_GIT_CLONE_CONNECTION_TYPE: LOOKER_BI_TYPE,
    }

    def __init__(self, config, command_name: str, **kwargs):
        super().__init__(config, command_name=command_name, **kwargs)

    @manage_errors
    def onboard_connection(self, connection_type: str, self_hosting_key: str, **kwargs) -> None:
        """
        Onboard a connection with self-hosted credentials by validating and adding a connection
        """
        kwargs["connectionType"] = connection_type

        try:
            region = AWSArn(self_hosting_key).region
        except IndexError:
            complain_and_abort("Credential key is not a valid ARN")

        if connection_type in CONNECTION_TO_WAREHOUSE_TYPE_MAP:
            kwargs["warehouseType"] = CONNECTION_TO_WAREHOUSE_TYPE_MAP[connection_type]
        elif connection_type in self.BI_CONNECTION_TYPES:
            kwargs["connection_query"] = ADD_BI_CONNECTION_MUTATION
            kwargs["connection_response"] = EXPECTED_ADD_BI_RESPONSE_FIELD
            kwargs["warehouse_type"] = self.BI_WAREHOUSE_TYPE_MAPPING.get(connection_type)

        self.onboard(
            validation_query=TEST_SELF_HOSTED_CRED_MUTATION,
            validation_response=EXPECTED_SELF_HOSTED_GQL_RESPONSE_FIELD,
            connection_type=connection_type,
            self_hosting_key=self_hosting_key,
            region=region,
            **kwargs,
        )

    @manage_errors
    def onboard_connection_v2(self, **kwargs) -> None:
        """
        Onboard a connection with self-hosted credentials using the v2 API.
        This method supports both environment variables and AWS Secrets Manager.
        """
        if kwargs.get("env_var_name") and not kwargs["env_var_name"].startswith("MCD_"):
            complain_and_abort("Environment variable name must start with prefix 'MCD_'")

        connection_type = str(kwargs["connection_type"])
        if connection_type in CONNECTION_TO_WAREHOUSE_TYPE_MAP:
            kwargs["warehouse_type"] = CONNECTION_TO_WAREHOUSE_TYPE_MAP[connection_type]
        elif connection_type in self.BI_CONNECTION_TYPES:
            kwargs["connection_query"] = ADD_BI_CONNECTION_MUTATION
            kwargs["connection_response"] = EXPECTED_ADD_BI_RESPONSE_FIELD
            kwargs["warehouse_type"] = self.BI_WAREHOUSE_TYPE_MAPPING.get(connection_type)

        # Remove None values
        connection_details = {k: v for k, v in kwargs.items() if v is not None}
        self.onboard(
            validation_query=TEST_SELF_HOSTED_CRED_V2_MUTATION,
            validation_response=EXPECTED_SELF_HOSTED_GQL_V2_RESPONSE_FIELD,
            connectionType=connection_type,
            validationName="save_credentials",
            warehouseName=kwargs["name"],
            # set connection to disabled upon creation
            is_active=False,
            **connection_details,
        )



import copy
import json
from unittest import TestCase, mock
from unittest.mock import Mock, patch

import click
from box import Box

from montecarlodata.common.data import (
    ConnectionOptions,
    MonolithResponse,
    OnboardingConfiguration,
    ValidationResult,
)
from montecarlodata.common.user import UserService
from montecarlodata.integrations.onboarding.base import BaseOnboardingService
from montecarlodata.integrations.onboarding.fields import (
    DATA_LAKE_WAREHOUSE_TYPE,
    EXPECTED_ADD_ETL_CONNECTION_RESPONSE_FIELD,
    EXPECTED_FIVETRAN_RESPONSE_FIELD,
)
from montecarlodata.queries.onboarding import (
    ADD_CONNECTION_MUTATION,
    ADD_ETL_CONNECTION_MUTATION,
    TEST_FIVETRAN_CRED_MUTATION,
)
from montecarlodata.utils import AwsClientWrapper, GqlWrapper
from tests.test_common_user import _SAMPLE_CONFIG, _SAMPLE_DW_ID, _SAMPLE_USER_RESPONSE

_SAMPLE_BUCKET_NAME = "bucket"
_SAMPLE_BASE_OPTIONS = {"foo": "bar"}
_SAMPLE_DC_OUTPUT = [
    {
        "OutputKey": "PrivateS3BucketArn",
        "OutputValue": f"arn:aws:s3:::{_SAMPLE_BUCKET_NAME}",
    }
]
_SAMPLE_CONNECTION_PATH = "test"


class BaseOnboardingTest(TestCase):
    def setUp(self) -> None:
        self._user_service_mock = Mock(autospec=UserService)
        self._request_wrapper_mock = Mock(autospec=GqlWrapper)
        self._aws_wrapper_mock = Mock(autospec=AwsClientWrapper)

        self._user_service_mock.collectors = _SAMPLE_USER_RESPONSE["getUser"]["account"][
            "dataCollectors"
        ]
        self._user_service_mock.active_collector = self._user_service_mock.collectors[0]
        self._user_service_mock.get_collector.return_value = Box(
            self._user_service_mock.collectors[0]
        )
        self._user_service_mock.get_collector_agent.return_value = None
        self._user_service_mock.warehouses = [
            {"uuid": _SAMPLE_DW_ID, "connectionType": "data-lake", "name": "data-lake"}
        ]

        self._service = BaseOnboardingService(
            _SAMPLE_CONFIG,
            command_name="test",
            request_wrapper=self._request_wrapper_mock,
            aws_wrapper=self._aws_wrapper_mock,
            user_service=self._user_service_mock,
        )

    @patch.object(BaseOnboardingService, "_validate_connection")
    @patch.object(BaseOnboardingService, "_add_connection")
    @patch("montecarlodata.integrations.onboarding.base.prompt_connection")
    def test_onboard_wrapper(self, prompt_mock, connection_mock, validation_mock):
        # Test the convenience wrapper makes the expected validation and connection calls
        validation_query, validation_response, connection_type, job_types = (
            "foo",
            "bar",
            "qux",
            ["test"],
        )
        options, v_return = (
            {"hello": "world"},
            ValidationResult(has_warnings=False, credentials_key="path"),
        )
        job_limits = {"tables_batch_size": 10}

        validation_mock.return_value = v_return

        self._service.onboard(
            validation_query=validation_query,
            validation_response=validation_response,
            connection_type=connection_type,
            job_types=job_types,
            job_limits=job_limits,
            **options,
        )

        validation_mock.assert_called_once_with(
            query=validation_query, response_field=validation_response, **options
        )

        expected_onboarding_config = OnboardingConfiguration(
            connection_type=connection_type,
            job_types=job_types,
            job_limits=job_limits,
            warehouse_type=DATA_LAKE_WAREHOUSE_TYPE,
            warehouse_name=connection_type,
            validation_query=validation_query,
            validation_response=validation_response,
            **{},
        )
        expected_onboarding_config.connection_options = ConnectionOptions(
            monolith_base_payload=options,
            dc_id=None,
            validate_only=False,
            skip_validation=False,
            skip_permission_tests=False,
            auto_yes=False,
        )

        connection_mock.assert_called_once_with(
            temp_path=v_return.credentials_key,
            onboarding_config=expected_onboarding_config,
        )
        prompt_mock.assert_called_once_with(
            message="Please confirm you want to add this connection", skip_prompt=False
        )

    @patch.object(BaseOnboardingService, "_validate_connection")
    @patch.object(BaseOnboardingService, "_add_connection")
    @patch("montecarlodata.integrations.onboarding.base.prompt_connection")
    def test_onboard_wrapper_with_options(self, prompt_mock, connection_mock, validation_mock):
        validation_query, validation_response, connection_type, job_types, dc_id = (
            "foo",
            "bar",
            "qux",
            ["test"],
            "123",
        )
        input_options = {
            "hello": "world",
            "dc_id": dc_id,
            "auto_yes": True,
            "skip_validation": True,
        }
        v_return = ValidationResult(has_warnings=False, credentials_key="path")

        monolith_connection_payload = {"dcId": "123", "skipValidation": True}

        validator_options = copy.deepcopy(input_options)
        validator_options["connectionOptions"] = {"dcId": dc_id, "skipValidation": True}
        [validator_options.pop(i) for i in ["dc_id", "auto_yes", "skip_validation"]]

        validation_mock.return_value = v_return

        self._service.onboard(
            validation_query=validation_query,
            validation_response=validation_response,
            connection_type=connection_type,
            job_types=job_types,
            **input_options,
        )

        validation_mock.assert_called_once_with(
            query=validation_query,
            response_field=validation_response,
            **validator_options,
        )

        expected_onboarding_config = OnboardingConfiguration(
            connection_type=connection_type,
            job_types=job_types,
            warehouse_type=DATA_LAKE_WAREHOUSE_TYPE,
            warehouse_name=connection_type,
            validation_query=validation_query,
            validation_response=validation_response,
            **{},
        )
        expected_onboarding_config.connection_options = ConnectionOptions(
            monolith_base_payload={
                "hello": "world",
                "connectionOptions": monolith_connection_payload,
            },
            dc_id=dc_id,
            validate_only=False,
            skip_validation=True,
            skip_permission_tests=False,
            auto_yes=True,
        )
        expected_onboarding_config.connection_options.monolith_connection_payload = (
            monolith_connection_payload
        )

        connection_mock.assert_called_once_with(
            temp_path=v_return.credentials_key,
            onboarding_config=expected_onboarding_config,
        )
        prompt_mock.assert_called_once_with(
            message="Please confirm you want to add this connection", skip_prompt=True
        )

    def test_handle_no_cert(self):
        original_options = copy.deepcopy(_SAMPLE_BASE_OPTIONS)
        self._service.handle_cert("test", _SAMPLE_BASE_OPTIONS)

        self.assertDictEqual(original_options, _SAMPLE_BASE_OPTIONS)
        self._aws_wrapper_mock.upload_file.assert_not_called()

    def test_handle_cert_in_s3(self):
        s3_key = "testing/test"
        options = {"cert_s3": s3_key, **_SAMPLE_BASE_OPTIONS}
        self._service.handle_cert("test", options)

        self.assertDictEqual(
            options,
            {
                "ssl_options": {"cert": s3_key, "mechanism": "dc-s3"},
                **_SAMPLE_BASE_OPTIONS,
            },
        )
        self._aws_wrapper_mock.upload_file.assert_not_called()

    def test_handle_cert_from_file(self):
        file_path = "./test/test.txt"
        s3_key = "test/test.txt"
        options = {"cert_file": file_path, **_SAMPLE_BASE_OPTIONS}

        self._aws_wrapper_mock.get_stack_outputs.return_value = _SAMPLE_DC_OUTPUT
        self._service.handle_cert("test", options)

        self._aws_wrapper_mock.upload_file.assert_called_once_with(
            bucket_name=_SAMPLE_BUCKET_NAME, object_name=s3_key, file_path=file_path
        )
        self.assertDictEqual(
            options,
            {
                "ssl_options": {"cert": s3_key, "mechanism": "dc-s3"},
                **_SAMPLE_BASE_OPTIONS,
            },
        )

    def test_handle_cert_from_file_agents(self):
        file_path = "./test/test.txt"
        options = {"cert_file": file_path, **_SAMPLE_BASE_OPTIONS}
        self._user_service_mock.get_collector_agent.return_value = Box(
            {"uuid": "foobar", "isDeleted": False}
        )
        with self.assertRaises(click.exceptions.Abort):
            self._service.handle_cert("test", options)

    @patch("montecarlodata.errors.click")
    def test_failed_validation(self, click_mock):
        click_mock.Abort.return_value = click.exceptions.Abort()
        expected_echo = "Error - Connection test failed!"

        query, response_field = "foo", "bar"
        self._request_wrapper_mock.make_request_v2.return_value = MonolithResponse(
            data={response_field: {}}
        )
        with self.assertRaises(click.exceptions.Abort):
            self._service._validate_connection(query, response_field, **_SAMPLE_BASE_OPTIONS)
        self._request_wrapper_mock.make_request_v2.assert_called_once_with(
            query=query,
            operation=response_field,
            service="onboarding_service",
            variables=_SAMPLE_BASE_OPTIONS,
        )

        click_mock.echo.assert_called_with(expected_echo, err=True)

    @patch("montecarlodata.errors.click")
    def test_failed_validation_with_details(self, click_mock):
        click_mock.Abort.return_value = click.exceptions.Abort()
        expected_echo = (
            "Error - Connection test failed!\n"
            "╒═══════════════╤══════════╤═════════════════════════════════════════════╕\n"
            "│ Validation    │ Result   │ Details                                     │\n"
            "╞═══════════════╪══════════╪═════════════════════════════════════════════╡\n"
            "│ Able to dance │ Passed   │                                             │\n"
            "├───────────────┼──────────┼─────────────────────────────────────────────┤\n"
            "│ Able to sing  │ Failed   │ Turns out I cannot carry a tune in a bucket │\n"
            "╘═══════════════╧══════════╧═════════════════════════════════════════════╛"
        )

        query, response_field = "foo", "bar"
        self._request_wrapper_mock.make_request_v2.return_value = MonolithResponse(
            data={
                "success": False,
                "validations": [{"type": "skills", "message": "Able to dance"}],
                "warnings": [
                    {
                        "type": "metadata",
                        "message": "Able to sing",
                        "data": {"error": "Turns out I cannot carry a tune in a bucket"},
                    }
                ],
            }
        )
        with self.assertRaises(click.exceptions.Abort):
            self._service._validate_connection(query, response_field, **_SAMPLE_BASE_OPTIONS)
        self._request_wrapper_mock.make_request_v2.assert_called_once_with(
            query=query,
            operation=response_field,
            service="onboarding_service",
            variables=_SAMPLE_BASE_OPTIONS,
        )

        click_mock.echo.assert_called_with(expected_echo, err=True)

    @patch("montecarlodata.integrations.onboarding.base.click")
    def test_successful_validation(self, click_mock):
        expected_echo = "Connection test was successful!"

        query, response_field, temp_path = "foo", "bar", "qux"
        self._request_wrapper_mock.make_request_v2.return_value = MonolithResponse(
            data={"key": temp_path}
        )

        result = self._service._validate_connection(query, response_field, **_SAMPLE_BASE_OPTIONS)
        self.assertFalse(result.has_warnings)
        self.assertEqual(temp_path, result.credentials_key)

        self._request_wrapper_mock.make_request_v2.assert_called_once_with(
            query=query,
            operation=response_field,
            service="onboarding_service",
            variables=_SAMPLE_BASE_OPTIONS,
        )

        click_mock.echo.assert_called_with(expected_echo)

    @patch("montecarlodata.integrations.onboarding.base.click")
    def test_successful_validation_with_warnings(self, click_mock):
        expected_echo = (
            "Connection test was successful!\n"
            "╒═══════════════╤══════════╤═════════════════════════════════════════════╕\n"
            "│ Validation    │ Result   │ Details                                     │\n"
            "╞═══════════════╪══════════╪═════════════════════════════════════════════╡\n"
            "│ Able to dance │ Passed   │                                             │\n"
            "├───────────────┼──────────┼─────────────────────────────────────────────┤\n"
            "│ Able to sing  │ Failed   │ Turns out I cannot carry a tune in a bucket │\n"
            "╘═══════════════╧══════════╧═════════════════════════════════════════════╛"
        )

        query, response_field, temp_path = "foo", "bar", "qux"
        self._request_wrapper_mock.make_request_v2.return_value = MonolithResponse(
            data={
                "success": True,
                "key": temp_path,
                "validations": [{"type": "skills", "message": "Able to dance"}],
                "warnings": [
                    {
                        "type": "metadata",
                        "message": "Able to sing",
                        "data": {"error": "Turns out I cannot carry a tune in a bucket"},
                    }
                ],
            }
        )

        result = self._service._validate_connection(query, response_field, **_SAMPLE_BASE_OPTIONS)
        self.assertTrue(result.has_warnings)
        self.assertEqual(temp_path, result.credentials_key)

        self._request_wrapper_mock.make_request_v2.assert_called_once_with(
            query=query,
            operation=response_field,
            service="onboarding_service",
            variables=_SAMPLE_BASE_OPTIONS,
        )

        click_mock.echo.assert_called_with(expected_echo)

    def test_connection_with_one_warehouse(self):
        key, conn_id = "foo", "bar"
        self._user_service_mock.warehouses = [
            {"uuid": _SAMPLE_DW_ID, "connectionType": "DATA_LAKE"}
        ]
        self._request_wrapper_mock.make_request_v2.return_value = MonolithResponse(
            data={"connection": {"uuid": conn_id}}
        )
        self.assertTrue(
            self._service._add_connection(
                key,
                OnboardingConfiguration(connection_type=_SAMPLE_CONNECTION_PATH, **{}),
            )
        )

        self._request_wrapper_mock.make_request_v2.assert_called_once_with(
            query=ADD_CONNECTION_MUTATION,
            operation="addConnection",
            service="onboarding_service",
            variables={
                "key": key,
                "connectionType": _SAMPLE_CONNECTION_PATH,
                "dwId": _SAMPLE_DW_ID,
            },
        )

    def test_connection_with_no_warehouses(self):
        key, conn_id = "foo", "bar"
        self._user_service_mock.warehouses = []
        service = BaseOnboardingService(
            _SAMPLE_CONFIG,
            command_name="test",
            request_wrapper=self._request_wrapper_mock,
            aws_wrapper=self._aws_wrapper_mock,
            user_service=self._user_service_mock,
        )
        self._request_wrapper_mock.make_request_v2.return_value = MonolithResponse(
            data={"connection": {"uuid": conn_id}}
        )
        self.assertTrue(
            service._add_connection(
                key,
                OnboardingConfiguration(connection_type="glue", warehouse_name="data-lake", **{}),
            )
        )

        self._request_wrapper_mock.make_request_v2.assert_called_once_with(
            query=ADD_CONNECTION_MUTATION,
            operation="addConnection",
            service="onboarding_service",
            variables={
                "key": key,
                "connectionType": "glue",
                "createWarehouseType": "data-lake",
                "name": "data-lake",
            },
        )

    def test_connection_with_multiple_warehouses(self):
        key, conn_id, name, dw_type = "foo", "bar", "test", "redshift"
        self._user_service_mock.warehouses = [
            {"uuid": "123", "connectionType": "redshift"},
            {"uuid": "456", "connectionType": "snowflake"},
        ]
        config = OnboardingConfiguration(connection_type=_SAMPLE_CONNECTION_PATH, **{})
        config.warehouse_type = dw_type
        config.warehouse_name = name

        self._request_wrapper_mock.make_request_v2.return_value = MonolithResponse(
            data={"connection": {"uuid": conn_id}}
        )
        self.assertTrue(self._service._add_connection(key, config))

        self._request_wrapper_mock.make_request_v2.assert_called_once_with(
            query=ADD_CONNECTION_MUTATION,
            operation="addConnection",
            service="onboarding_service",
            variables={
                "key": key,
                "connectionType": _SAMPLE_CONNECTION_PATH,
                "createWarehouseType": dw_type,
                "name": name,
            },
        )

    def test_connection_with_multiple_warehouses_and_a_lake(self):
        key, conn_id, dw_id = "foo", "bar", "456"
        self._user_service_mock.warehouses = [
            {"uuid": "123", "connectionType": "redshift", "name": "redshift"},
            {"uuid": dw_id, "connectionType": "data-lake", "name": "data-lake"},
        ]

        self._request_wrapper_mock.make_request_v2.return_value = MonolithResponse(
            data={"connection": {"uuid": conn_id}}
        )
        self.assertTrue(
            self._service._add_connection(
                key,
                OnboardingConfiguration(
                    connection_type=_SAMPLE_CONNECTION_PATH,
                    warehouse_name="data-lake",
                    **{},
                ),
            )
        )

        self._request_wrapper_mock.make_request_v2.assert_called_once_with(
            query=ADD_CONNECTION_MUTATION,
            operation="addConnection",
            service="onboarding_service",
            variables={
                "key": key,
                "connectionType": _SAMPLE_CONNECTION_PATH,
                "dwId": dw_id,
                "name": "data-lake",
            },
        )

    def test_connection_with_multiple_warehouses_adding_connection_to_existing_lake_warehouse(
        self,
    ):
        key, conn_id, dw_id = "foo", "bar", "456"
        self._user_service_mock.warehouses = [
            {"uuid": "123", "connectionType": "redshift", "name": "redshift"},
            {"uuid": dw_id, "connectionType": "data-lake", "name": "data-lake"},
        ]

        self._request_wrapper_mock.make_request_v2.return_value = MonolithResponse(
            data={"connection": {"uuid": conn_id}}
        )
        self.assertTrue(
            self._service._add_connection(
                key,
                OnboardingConfiguration(
                    connection_type="hive-mysql",
                    warehouse_name="data-lake",
                    create_warehouse=False,
                    **{},
                ),
            )
        )

        self._request_wrapper_mock.make_request_v2.assert_called_once_with(
            query=ADD_CONNECTION_MUTATION,
            operation="addConnection",
            service="onboarding_service",
            variables={
                "key": key,
                "connectionType": "hive-mysql",
                "dwId": dw_id,
                "name": "data-lake",
            },
        )

    def test_connection_with_multiple_warehouses_adding_lake_warehouse(self):
        key, conn_id, dw_id = "foo", "bar", "456"
        self._user_service_mock.warehouses = [
            {"uuid": "123", "connectionType": "redshift", "name": "redshift"},
            {"uuid": dw_id, "connectionType": "data-lake", "name": "data-lake"},
        ]

        self._request_wrapper_mock.make_request_v2.return_value = MonolithResponse(
            data={"connection": {"uuid": conn_id}}
        )
        self.assertTrue(
            self._service._add_connection(
                key,
                OnboardingConfiguration(
                    connection_type="hive-mysql",
                    warehouse_name="data-lake2",
                    create_warehouse=True,
                    **{},
                ),
            )
        )

        self._request_wrapper_mock.make_request_v2.assert_called_once_with(
            query=ADD_CONNECTION_MUTATION,
            operation="addConnection",
            service="onboarding_service",
            variables={
                "key": key,
                "connectionType": "hive-mysql",
                "createWarehouseType": "data-lake",
                "name": "data-lake2",
            },
        )

    def test_connection_with_job_types(self):
        key, conn_id, jobs = "foo", "bar", ["qux"]
        config = OnboardingConfiguration(connection_type=_SAMPLE_CONNECTION_PATH, **{})
        config.job_types = jobs

        self._request_wrapper_mock.make_request_v2.return_value = MonolithResponse(
            data={"connection": {"uuid": conn_id}}
        )
        self.assertTrue(self._service._add_connection(key, config))

        self._request_wrapper_mock.make_request_v2.assert_called_once_with(
            query=ADD_CONNECTION_MUTATION,
            operation="addConnection",
            service="onboarding_service",
            variables={
                "key": key,
                "connectionType": _SAMPLE_CONNECTION_PATH,
                "dwId": _SAMPLE_DW_ID,
                "jobTypes": jobs,
            },
        )

    def test_connection_with_job_limits(self):
        key, conn_id, job_limits = "foo", "bar", {"hello": "world"}
        config = OnboardingConfiguration(connection_type=_SAMPLE_CONNECTION_PATH, **{})
        config.job_limits = job_limits

        self._request_wrapper_mock.make_request_v2.return_value = MonolithResponse(
            data={"connection": {"uuid": conn_id}}
        )
        self.assertTrue(self._service._add_connection(key, config))

        self._request_wrapper_mock.make_request_v2.assert_called_once_with(
            query=ADD_CONNECTION_MUTATION,
            operation="addConnection",
            service="onboarding_service",
            variables={
                "key": key,
                "connectionType": _SAMPLE_CONNECTION_PATH,
                "dwId": _SAMPLE_DW_ID,
                "jobLimits": json.dumps(job_limits),
            },
        )

    def test_connection_for_etl(self):
        connection_type = "fivetran"
        dc_id = "dc_id"
        connection_name = "FIVETRAN_TEST_DB"
        fivetran_key = "fivetran_key"
        fivetran_password = "fivetran_password"
        fivetran_url = "https://api.fivetran.com/v1/"

        kwargs = {
            "validation_query": TEST_FIVETRAN_CRED_MUTATION,
            "validation_response": EXPECTED_FIVETRAN_RESPONSE_FIELD,
            "connection_query": ADD_ETL_CONNECTION_MUTATION,
            "connection_response": EXPECTED_ADD_ETL_CONNECTION_RESPONSE_FIELD,
            "connection_type": connection_type,
            "warehouse_name": connection_name,
            "fivetran_api_key": fivetran_key,
            "fivetran_api_password": fivetran_password,
            "fivetran_base_url": fivetran_url,
            "dc_id": dc_id,
            "skip_validation": False,
            "validate_only": False,
            "auto_yes": True,
        }

        connection_validation_response_key = "tmp/8d1e5786-5afb-46c5-b297-3fc355504bb2"

        monolith_base_payload = {
            "fivetran_api_key": fivetran_key,
            "fivetran_api_password": fivetran_password,
            "fivetran_base_url": fivetran_url,
            "connectionOptions": {"dcId": dc_id},
        }

        connection_request = {
            "key": connection_validation_response_key,
            "connectionType": connection_type,
            "name": connection_name,
            "dcId": dc_id,
        }

        self._request_wrapper_mock.make_request_v2.side_effect = [
            MonolithResponse(data={"key": connection_validation_response_key, "success": True}),
            MonolithResponse(data={"connection": {"uuid": "conn_id"}}),
        ]

        self._request_wrapper_mock.make_request_v2.return_value = MonolithResponse(
            data={"key": connection_validation_response_key, "success": True}
        )

        self._service.onboard(**kwargs)

        self.assertEqual(
            self._request_wrapper_mock.make_request_v2.mock_calls,
            [
                mock.call(
                    query=TEST_FIVETRAN_CRED_MUTATION,
                    operation=EXPECTED_FIVETRAN_RESPONSE_FIELD,
                    service="onboarding_service",
                    variables=monolith_base_payload,
                ),
                mock.call(
                    query=ADD_ETL_CONNECTION_MUTATION,
                    operation=EXPECTED_ADD_ETL_CONNECTION_RESPONSE_FIELD,
                    service="onboarding_service",
                    variables=connection_request,
                ),
            ],
        )



from montecarlodata.config import Config
from montecarlodata.errors import manage_errors
from montecarlodata.integrations.onboarding.base import BaseOnboardingService
from montecarlodata.integrations.onboarding.fields import (
    EXPECTED_ADD_ETL_CONNECTION_RESPONSE_FIELD,
    EXPECTED_FIVETRAN_RESPONSE_FIELD,
    FIVETRAN_CONNECTION_TYPE,
)
from montecarlodata.queries.onboarding import (
    ADD_ETL_CONNECTION_MUTATION,
    TEST_FIVETRAN_CRED_MUTATION,
)


class FivetranOnboardingService(BaseOnboardingService):
    def __init__(self, config: Config, command_name: str, **kwargs):
        super().__init__(config, command_name=command_name, **kwargs)

    @manage_errors
    def onboard_fivetran(self, **kwargs) -> None:
        self.onboard(
            validation_query=TEST_FIVETRAN_CRED_MUTATION,
            validation_response=EXPECTED_FIVETRAN_RESPONSE_FIELD,
            connection_query=ADD_ETL_CONNECTION_MUTATION,
            connection_response=EXPECTED_ADD_ETL_CONNECTION_RESPONSE_FIELD,
            connection_type=FIVETRAN_CONNECTION_TYPE,
            **kwargs,
        )



import click

_FG_COLOR_SUCCESS = "green"
_FG_COLOR_ERROR = "red"
_FG_COLOR_WARNING = "yellow"


def echo_success_message(message: str):
    click.echo(click.style(message, fg=_FG_COLOR_SUCCESS))


def echo_error_message(message: str):
    click.echo(click.style("Error: ", fg=_FG_COLOR_ERROR) + message)


def echo_warning_message(message: str):
    click.echo(click.style("Warn: ", fg=_FG_COLOR_WARNING) + message)


def styled_success_icon() -> str:
    return click.style("\u2713", fg=_FG_COLOR_SUCCESS)


def styled_error_icon() -> str:
    return click.style("\u2715", fg=_FG_COLOR_ERROR)


def styled_warning_icon() -> str:
    return click.style("\u26a0", fg=_FG_COLOR_WARNING)



from typing import Iterator, List, Optional, Set, Tuple
from urllib.parse import urlparse

import click
from box import Box
from click._compat import get_text_stderr

from montecarlodata.common.common import generate_token, read_files, render_dumped_json
from montecarlodata.common.data import (
    AWSArn,
    AwsGlueAthenaResourceProperties,
    DcResourceProperties,
)
from montecarlodata.common.resources import CloudResourceService
from montecarlodata.common.user import UserService
from montecarlodata.config import Config
from montecarlodata.errors import complain_and_abort, manage_errors
from montecarlodata.utils import GqlWrapper


class PolicyDiscoveryService:
    ANY_RESOURCE_KEY = "*"

    # Rel template paths. Full path retrieved via pkgutil.
    CF_ROLE_REL_PATH = "../templates/aws_role.json"
    GLUE_POLICY_REL_PATH = "../templates/glue_policy.json"
    ATHENA_POLICY_REL_PATH = "../templates/athena_policy.json"
    MSK_POLICY_REL_PATH = "../templates/msk_policy.json"

    def __init__(
        self,
        config: Config,
        command_name: str,
        cloud_resource_service: Optional[CloudResourceService] = None,
        user_service: Optional[UserService] = None,
        request_wrapper: Optional[GqlWrapper] = None,
        *args,
        **kwargs,
    ):
        self._abort_on_error = True  # Used in decorator
        self._command_name = command_name
        self._cloud_resource_service = cloud_resource_service or CloudResourceService(
            *args,
            config=config,
            command_name=self._command_name,
            **kwargs,
        )
        self._request_wrapper = request_wrapper or GqlWrapper(
            config,
            command_name=self._command_name,
        )
        self._user_service = user_service or UserService(
            config=config,
            request_wrapper=self._request_wrapper,
            command_name=self._command_name,
        )

    @manage_errors
    def generate_cf_role(
        self,
        policy_files: List,
        dc_id: Optional[str] = None,
        agent_id: Optional[str] = None,
    ) -> None:
        """
        Generate a CloudFormation template to create a MC compatible IAM role from one
        or more IAM policy.
        """
        click.echo("Generating CloudFormation template for creating a role resource.", err=True)
        if not len(policy_files) > 0:
            complain_and_abort("At least one policy is required.")

        collector_arn = AWSArn(
            self._cloud_resource_service.get_and_validate_active_collector(
                agent_id=agent_id, dc_id=dc_id
            ).stackArn
        )

        click.echo(
            render_dumped_json(
                path=self.CF_ROLE_REL_PATH,
                collector_account_id=collector_arn.account,
                collector_external_id=generate_token(),
                resource_policies=read_files(policy_files),
            )
        )

    @manage_errors
    def generate_msk_policy(self, **kwargs) -> None:
        """
        Generate an IAM compatible policy for MSK.
        """
        click.echo("Generating policy for MSK.", err=True)

        dc_resource_props = self._cloud_resource_service.get_dc_resource_props(
            collector_props=self._cloud_resource_service.get_and_validate_active_collector(
                agent_id=kwargs.get("agent_id"), dc_id=kwargs.get("dc_id")
            ),
            get_stack_outputs=False,
            get_stack_params=False,
            **kwargs,
        )

        click.echo(
            render_dumped_json(
                path=self.MSK_POLICY_REL_PATH,
                account_id=dc_resource_props.resources_client.get_caller_identity(),
                region=dc_resource_props.resources_region,
            )
        )

    @manage_errors
    def generate_glue_policy(self, **kwargs) -> None:
        """
        Generate an IAM compatible policy for Glue from the specified database names or wildcard
        sequences.

        Bucket names are derived from the database names if not otherwise provided.
        """
        click.echo("Generating policy for Glue.", err=True)
        glue_properties = self._get_common_glue_athena_props(**kwargs)

        assert glue_properties.dc_resource_props is not None, "DC resource properties not found"
        click.echo(
            render_dumped_json(
                path=self.GLUE_POLICY_REL_PATH,
                account_id=glue_properties.dc_resource_props.resources_client.get_caller_identity(),
                region=glue_properties.dc_resource_props.resources_region,
                database_names=glue_properties.database_names,
                data_buckets=glue_properties.bucket_names,
            )
        )

    @manage_errors
    def generate_athena_policy(self, *, workgroup_name: str, **kwargs) -> None:
        """
        Generate an IAM compatible policy for Athena from the specified database names or wildcard
        sequences.

        Bucket names are derived from the database names if not otherwise provided. The results
        bucket is derived from the workgroup.
        """
        click.echo("Generating policy for Athena.", err=True)
        athena_properties = self._get_common_glue_athena_props(**kwargs)

        assert athena_properties.dc_resource_props is not None, "DC resource properties not found"
        result_bucket, result_path = self._get_workgroup_results(
            workgroup_name=workgroup_name,
            dc_resource_props=athena_properties.dc_resource_props,
        )

        click.echo(
            render_dumped_json(
                path=self.ATHENA_POLICY_REL_PATH,
                account_id=athena_properties.dc_resource_props.resources_client.get_caller_identity(),
                region=athena_properties.dc_resource_props.resources_region,
                database_names=athena_properties.database_names,
                data_buckets=athena_properties.bucket_names,
                result_bucket=result_bucket,
                result_path=result_path,
                workgroup_name=workgroup_name,
            )
        )

    def _get_common_glue_athena_props(
        self,
        *,
        database_names: List[str],
        bucket_names: Optional[List[str]] = None,
        dc_id: Optional[str] = None,
        agent_id: Optional[str] = None,
        **kwargs,
    ) -> AwsGlueAthenaResourceProperties:
        """
        Get resources that are common to Glue and Athena.
        """
        bucket_names = bucket_names or []
        database_names_set = self._handle_wildcards(database_names)
        bucket_names_set = self._handle_wildcards(bucket_names)

        if agent_id and not dc_id:
            agent_dict = self._user_service.get_agent(agent_id=agent_id)
            dc_id = agent_dict["dc_id"]  # type: ignore

        dc_resource_props = self._cloud_resource_service.get_dc_resource_props(
            collector_props=self._cloud_resource_service.get_and_validate_active_collector(
                dc_id=dc_id
            ),
            get_stack_outputs=False,
            get_stack_params=False,
            **kwargs,
        )

        if not bucket_names_set:
            bucket_names_set = self._get_buckets(
                database_names=database_names_set, dc_resource_props=dc_resource_props
            )

        return AwsGlueAthenaResourceProperties(
            database_names=database_names_set,
            bucket_names=bucket_names_set,
            dc_resource_props=dc_resource_props,
        )

    def _get_buckets(
        self, database_names: Set[str], dc_resource_props: DcResourceProperties
    ) -> Set[str]:
        """
        Derive s3 locations (buckets) from provided databases.
        """
        if self.ANY_RESOURCE_KEY in database_names:
            return {self.ANY_RESOURCE_KEY}

        buckets = set()
        with click.progressbar(
            database_names, label="Looking up buckets details.", file=get_text_stderr()
        ) as dbs:
            for database_name in dbs:
                page_token = None
                while True:
                    tables = dc_resource_props.resources_client.get_glue_tables(
                        database_name=database_name, page_token=page_token
                    )
                    page_token = tables.next_token
                    buckets.update(self._get_table_locations(tables=tables.table_list))

                    if not page_token:
                        break
        return buckets

    def _handle_wildcards(self, sequence: List[str]) -> Set[str]:
        """
        Get unique records and prioritize any wildcards in the sequence.
        """
        sequence_set = set(sequence) if sequence else set()
        if self.ANY_RESOURCE_KEY in sequence_set:
            return {self.ANY_RESOURCE_KEY}
        return sequence_set

    @staticmethod
    def _get_workgroup_results(
        workgroup_name: str, dc_resource_props: DcResourceProperties
    ) -> Tuple[str, str]:
        """
        Get workgroup result bucket and path.
        """
        with click.progressbar(
            length=1, label="Looking up workgroup details.", file=get_text_stderr()
        ) as progress:
            workgroup_details = dc_resource_props.resources_client.get_athena_workgroup(
                workgroup=workgroup_name
            )

            url = urlparse(workgroup_details.configuration.result_configuration.output_location)
            progress.update(1)
            return url.netloc, url.path if url.path else "/"

    @staticmethod
    def _get_table_locations(
        tables: Box,
        valid_scheme_prefix: str = "s3",
    ) -> Iterator[str]:
        """
        Get locations from s3 for any valid schemes.
        """
        for table in tables:
            url = urlparse(table.storage_descriptor.location)
            if url.scheme.lower().startswith(valid_scheme_prefix):
                yield url.netloc



from montecarlodata.queries.common import GQL


class Queries:
    create = GQL(
        query="""
        mutation ($description: String!, $scope: IntegrationKeyScope!, $warehouseIds: [UUID]) {
            createIntegrationKey(description: $description, scope: $scope, warehouseIds: $warehouseIds) {
                key {
                    id
                    secret
                }
            }
        }
        """,
        operation="createIntegrationKey",
    )

    delete = GQL(
        query="""
        mutation ($keyId: String!) {
            deleteIntegrationKey(keyId: $keyId) {
                deleted
            }
        }
        """,
        operation="deleteIntegrationKey",
    )

    get_all = GQL(
        query="""
        query ($scope: String, $resourceUuid: UUID) {
            getIntegrationKeys(scope: $scope, resourceUuid: $resourceUuid) {
                id
                description
                scope
                createdTime,
                createdBy {
                    id,
                    firstName,
                    lastName,
                    email
                }
            }
        }
        """,
        operation="getIntegrationKeys",
    )



from dataclasses import dataclass


@dataclass
class DataExport:
    name: str
    title: str
    description: str


DATA_EXPORT_CATALOG = [
    DataExport(
        name="MONITORS",
        title="Monitors",
        description="All monitors with aggregated properties, excluding deleted monitors.",
    ),
    DataExport(
        name="ASSETS",
        title="Assets",
        description="All assets with aggregated properties, excluding deleted assets.",
    ),
    DataExport(
        name="ALERTS",
        title="Alerts",
        description="All alerts in the last 90 days with aggregated properties.",
    ),
    DataExport(
        name="EVENTS",
        title="Events",
        description="All events in the last 90 days with aggregated properties.",
    ),
]
